"""Defines a pipeline is exposed to the user, accessible via pipeline."""

# pylint: disable=too-many-lines

from __future__ import annotations

from copy import deepcopy
from typing import Any, Iterable, List, Literal, Optional, Tuple, TypeVar, Union

try:
    from typing import Self  # type: ignore[attr-defined]
except ImportError:
    from typing_extensions import Self

import joblib
import numpy as np
import numpy.typing as npt
from loguru import logger
from sklearn.base import _fit_context, clone
from sklearn.pipeline import Pipeline as _Pipeline
from sklearn.pipeline import _final_estimator_has, _fit_transform_one
from sklearn.utils import Bunch
from sklearn.utils._tags import Tags, get_tags
from sklearn.utils.metadata_routing import (
    MetadataRouter,
    MethodMapping,
    _routing_enabled,
    process_routing,
)
from sklearn.utils.metaestimators import available_if
from sklearn.utils.validation import check_memory

from molpipeline.abstract_pipeline_elements.core import ABCPipelineElement
from molpipeline.error_handling import ErrorFilter, FilterReinserter
from molpipeline.pipeline._molpipeline import _MolPipeline
from molpipeline.post_prediction import (
    PostPredictionTransformation,
    PostPredictionWrapper,
)
from molpipeline.utils.logging import print_elapsed_time
from molpipeline.utils.molpipeline_types import (
    AnyElement,
    AnyPredictor,
    AnyStep,
    AnyTransformer,
)
from molpipeline.utils.value_checks import is_empty

__all__ = ["Pipeline"]

# Type definitions
_T = TypeVar("_T")
# Cannot be moved to utils.molpipeline_types due to circular imports


_IndexedStep = Tuple[int, str, AnyElement]
_AggStep = Tuple[List[int], List[str], _MolPipeline]
_AggregatedPipelineStep = Union[_IndexedStep, _AggStep]


class Pipeline(_Pipeline):
    """Defines the pipeline which handles pipeline elements."""

    steps: list[AnyStep]
    #  * Adapted methods from sklearn.pipeline.Pipeline *

    def __init__(
        self,
        steps: list[AnyStep],
        *,
        memory: Optional[Union[str, joblib.Memory]] = None,
        verbose: bool = False,
        n_jobs: int = 1,
    ):
        """Initialize Pipeline.

        Parameters
        ----------
        steps: list[tuple[str, Union[AnyTransformer, AnyPredictor, ABCPipelineElement]]]
            List of (name, Estimator) tuples.
        memory: str, optional
            Path to cache transformers.
        verbose: bool, optional
            If True, print additional information.
        n_jobs: int, optional
            Number of cores used for aggregated steps.
        """
        super().__init__(steps, memory=memory, verbose=verbose)
        self.n_jobs = n_jobs
        self._set_error_resinserter()

    def _set_error_resinserter(self) -> None:
        """Connect the error resinserters with the error filters."""
        error_replacer_list = [
            e_filler
            for _, e_filler in self.steps
            if isinstance(e_filler, FilterReinserter)
        ]
        error_filter_list = [
            n_filter for _, n_filter in self.steps if isinstance(n_filter, ErrorFilter)
        ]
        for step in self.steps:
            if isinstance(step[1], PostPredictionWrapper):
                if isinstance(step[1].wrapped_estimator, FilterReinserter):
                    error_replacer_list.append(step[1].wrapped_estimator)
        for error_replacer in error_replacer_list:
            error_replacer.select_error_filter(error_filter_list)

    def _validate_steps(self) -> None:
        """Validate the steps."""
        names = [name for name, _ in self.steps]

        # validate names
        self._validate_names(names)

        # validate estimators
        non_post_processing_steps = [e for _, _, e in self._agg_non_postpred_steps()]
        transformer_list = non_post_processing_steps[:-1]
        estimator = non_post_processing_steps[-1]

        for transformer in transformer_list:
            if transformer is None or transformer == "passthrough":
                continue
            if not (
                hasattr(transformer, "fit") or hasattr(transformer, "fit_transform")
            ) or not hasattr(transformer, "transform"):
                raise TypeError(
                    f"All intermediate steps should be "
                    f"transformers and implement fit and transform "
                    f"or be the string 'passthrough' "
                    f"'{transformer}' (type {type(transformer)}) doesn't"
                )

        # We allow last estimator to be None as an identity transformation
        if (
            estimator is not None
            and estimator != "passthrough"
            and not hasattr(estimator, "fit")
        ):
            raise TypeError(
                f"Last step of Pipeline should implement fit "
                f"or be the string 'passthrough'. "
                f"'{estimator}' (type {type(estimator)}) doesn't"
            )

        # validate post-processing steps
        # Calling steps automatically validates them
        _ = self._post_processing_steps()

    def _iter(
        self, with_final: bool = True, filter_passthrough: bool = True
    ) -> Iterable[_AggregatedPipelineStep]:
        """Iterate over all non post-processing steps.

        Steps which are children of a ABCPipelineElement were aggregated to a MolPipeline.

        Parameters
        ----------
        with_final: bool, optional
            If True, the final estimator is included.
        filter_passthrough: bool, optional
            If True, passthrough steps are filtered out.

        Returns
        ------
        Iterable[_AggregatedPipelineStep]
            The _AggregatedPipelineStep is composed of the index, the name and the transformer.
        """
        last_element: Optional[_AggregatedPipelineStep] = None

        # This loop delays the output by one in order to identify the last step
        for step in self._agg_non_postpred_steps():
            # Only happens for the first step
            if last_element is None:
                last_element = step
                continue
            if not filter_passthrough:
                yield last_element
            elif step[2] is not None and step[2] != "passthrough":
                yield last_element
            last_element = step

        # This can only happen if no steps are set.
        if last_element is None:
            raise AssertionError("Pipeline needs to have at least one step!")

        if with_final and last_element[2] is not None:
            if last_element[2] != "passthrough":
                yield last_element

    @property
    def _estimator_type(self) -> Any:
        """Return the estimator type."""
        if self._final_estimator is None or self._final_estimator == "passthrough":
            return None
        if hasattr(self._final_estimator, "_estimator_type"):
            # pylint: disable=protected-access
            return self._final_estimator._estimator_type
        return None

    @property
    def _final_estimator(
        self,
    ) -> Union[
        Literal["passthrough"],
        AnyTransformer,
        AnyPredictor,
        _MolPipeline,
        ABCPipelineElement,
    ]:
        """Return the lst estimator which is not a PostprocessingTransformer."""
        element_list = list(self._agg_non_postpred_steps())
        last_element = element_list[-1]
        return last_element[2]

    # pylint: disable=too-many-locals,too-many-branches
    def _fit(
        self,
        X: Any,
        y: Any = None,
        routed_params: dict[str, Any] | None = None,
        raw_params: dict[str, Any] | None = None,
    ) -> tuple[Any, Any]:
        """Fit the model by fitting all transformers except the final estimator.

        Data can be subsetted by the transformers.

        Parameters
        ----------
        X : Any
            Training data.
        y : Any, optional (default=None)
            Training objectives.
        routed_params : dict[str, Any], optional
            Parameters for each step as returned by process_routing.
            Although this is marked as optional, it should not be None.
            The awkward (argward?) typing is due to inheritance from sklearn.
            Can be an empty dictionary.
        raw_params : dict[str, Any], optional
            Parameters passed by the user, used when `transform_input`

        Returns
        -------
        tuple[Any, Any]
            The transformed data and the transformed objectives.
        """
        # shallow copy of steps - this should really be steps_
        self.steps = list(self.steps)
        self._validate_steps()
        if routed_params is None:
            raise AssertionError("routed_params should not be None.")

        # Set up the memory
        memory: joblib.Memory = check_memory(self.memory)

        fit_transform_one_cached = memory.cache(_fit_transform_one)
        for step in self._iter(with_final=False, filter_passthrough=False):
            step_idx, name, transformer = step
            if transformer is None or transformer == "passthrough":
                with print_elapsed_time("Pipeline", self._log_message(step_idx)):
                    continue

            if hasattr(memory, "location") and memory.location is None:
                # we do not clone when caching is disabled to
                # preserve backward compatibility
                cloned_transformer = transformer
            else:
                cloned_transformer = clone(transformer)
            if isinstance(cloned_transformer, _MolPipeline):
                if routed_params:
                    step_params = {
                        "element_parameters": [routed_params[n] for n in name]
                    }
                else:
                    step_params = {}
            elif isinstance(name, list):
                raise AssertionError()
            else:
                step_params = self._get_metadata_for_step(
                    step_idx=step_idx,
                    step_params=routed_params[name],
                    all_params=raw_params,
                )

            # Fit or load from cache the current transformer
            X, fitted_transformer = fit_transform_one_cached(
                cloned_transformer,
                X,
                y,
                None,
                message_clsname="Pipeline",
                message=self._log_message(step_idx),
                params=step_params,
            )
            # Replace the transformer of the step with the fitted
            # transformer. This is necessary when loading the transformer
            # from the cache.
            if isinstance(fitted_transformer, _MolPipeline):
                ele_list = fitted_transformer.element_list
                if not isinstance(name, list) or not isinstance(step_idx, list):
                    raise AssertionError()
                if not len(name) == len(step_idx) == len(ele_list):
                    raise AssertionError()
                for idx_i, name_i, ele_i in zip(step_idx, name, ele_list):
                    self.steps[idx_i] = (name_i, ele_i)
                if y is not None:
                    y = fitted_transformer.co_transform(y)
                for idx_i, name_i, ele_i in zip(step_idx, name, ele_list):
                    self.steps[idx_i] = (name_i, ele_i)
                self._set_error_resinserter()
            elif isinstance(name, list) or isinstance(step_idx, list):
                raise AssertionError()
            else:
                self.steps[step_idx] = (name, fitted_transformer)
            if is_empty(X):
                return np.array([]), np.array([])
        return X, y

    def _transform(
        self, X: Any, routed_params: Bunch  # pylint: disable=invalid-name
    ) -> Any:
        """Transform the data, and skip final estimator.

        Call `transform` of each transformer in the pipeline except the last one,

        Parameters
        ----------
        X : iterable
            Data to predict on. Must fulfill input requirements of first step
            of the pipeline.

        routed_params: Bunch
            parameters for each step as returned by process_routing

        Returns
        -------
        Any
            Result of calling `transform` on the second last estimator.
        """
        iter_input = X
        do_routing = _routing_enabled()
        if do_routing:
            logger.warning("Routing is enabled and NOT fully tested!")

        for _, name, transform in self._iter(with_final=False):
            if is_empty(iter_input):
                if isinstance(transform, _MolPipeline):
                    _ = transform.transform(iter_input)
                iter_input = []
                break
            if transform == "passthrough":
                raise AssertionError("Passthrough should have been filtered out.")
            if hasattr(transform, "transform"):
                if do_routing:
                    iter_input = transform.transform(  # type: ignore[call-arg]
                        iter_input, routed_params[name].transform
                    )
                else:
                    iter_input = transform.transform(iter_input)
            else:
                raise AssertionError(
                    f"Non transformer ocurred in transformation step: {transform}."
                )
        return iter_input

    # * New implemented methods *
    def _non_post_processing_steps(
        self,
    ) -> list[AnyStep]:
        """Return all steps before the first PostPredictionTransformation.

        Returns
        -------
        list[AnyStep]
            List of steps before the first PostPredictionTransformation.
        """
        non_post_processing_steps: list[AnyStep] = []
        start_adding = False
        for step_name, step_estimator in self.steps[::-1]:
            if not isinstance(step_estimator, PostPredictionTransformation):
                start_adding = True
            if start_adding:
                if isinstance(step_estimator, PostPredictionTransformation):
                    raise AssertionError(
                        "PipelineElement of type PostPredictionTransformation occured before the last step."
                    )
                non_post_processing_steps.append((step_name, step_estimator))
        return list(non_post_processing_steps[::-1])

    def _post_processing_steps(self) -> list[tuple[str, PostPredictionTransformation]]:
        """Return last steps which are PostPredictionTransformation.

        Returns
        -------
        list[tuple[str, PostPredictionTransformation]]
            List of tuples containing the name and the PostPredictionTransformation.
        """
        post_processing_steps = []
        for step_name, step_estimator in self.steps[::-1]:
            if isinstance(step_estimator, PostPredictionTransformation):
                post_processing_steps.append((step_name, step_estimator))
            else:
                break
        return list(post_processing_steps[::-1])

    def _agg_non_postpred_steps(
        self,
    ) -> Iterable[_AggregatedPipelineStep]:
        """Generate (idx, (name, trans)) tuples from self.steps.

        When filter_passthrough is True, 'passthrough' and None transformers
        are filtered out.

        Returns
        ------
        Iterable[_AggregatedPipelineStep]
            The _AggregatedPipelineStep is composed of the index, the name and the transformer.
        """
        aggregated_transformer_list = []
        for i, (name_i, step_i) in enumerate(self._non_post_processing_steps()):
            if isinstance(step_i, ABCPipelineElement):
                aggregated_transformer_list.append((i, name_i, step_i))
            else:
                if aggregated_transformer_list:
                    index_list = [step[0] for step in aggregated_transformer_list]
                    name_list = [step[1] for step in aggregated_transformer_list]
                    transformer_list = [step[2] for step in aggregated_transformer_list]
                    if len(aggregated_transformer_list) == 1:
                        yield index_list[0], name_list[0], transformer_list[0]
                    else:
                        pipeline = _MolPipeline(transformer_list, n_jobs=self.n_jobs)
                        yield index_list, name_list, pipeline
                    aggregated_transformer_list = []
                yield i, name_i, step_i

        # yield last step if anything remains
        if aggregated_transformer_list:
            index_list = [step[0] for step in aggregated_transformer_list]
            name_list = [step[1] for step in aggregated_transformer_list]
            transformer_list = [step[2] for step in aggregated_transformer_list]

            if len(aggregated_transformer_list) == 1:
                yield index_list[0], name_list[0], transformer_list[0]

            elif len(aggregated_transformer_list) > 1:
                pipeline = _MolPipeline(transformer_list, n_jobs=self.n_jobs)
                yield index_list, name_list, pipeline

    @_fit_context(
        # estimators in Pipeline.steps are not validated yet
        prefer_skip_nested_validation=False
    )
    def fit(self, X: Any, y: Any = None, **fit_params: Any) -> Self:
        """Fit the model.

        Fit all the transformers one after the other and transform the
        data. Finally, fit the transformed data using the final estimator.

        Parameters
        ----------
        X : iterable
            Training data. Must fulfill input requirements of first step of the
            pipeline.

        y : iterable, default=None
            Training targets. Must fulfill label requirements for all steps of
            the pipeline.

        **fit_params : dict of string -> object
            Parameters passed to the ``fit`` method of each step, where
            each parameter name is prefixed such that parameter ``p`` for step
            ``s`` has key ``s__p``.

        Returns
        -------
        self : object
            Pipeline with fitted steps.
        """
        routed_params = self._check_method_params(method="fit", props=fit_params)
        Xt, yt = self._fit(X, y, routed_params)  # pylint: disable=invalid-name
        with print_elapsed_time("Pipeline", self._log_message(len(self.steps) - 1)):
            if self._final_estimator != "passthrough":
                if is_empty(Xt):
                    logger.warning(
                        "All input rows were filtered out! Model is not fitted!"
                    )
                else:
                    fit_params_last_step = routed_params[
                        self._non_post_processing_steps()[-1][0]
                    ]
                    self._final_estimator.fit(Xt, yt, **fit_params_last_step["fit"])

        return self

    def _can_fit_transform(self) -> bool:
        """Check if the final estimator can fit_transform or is passthrough.

        Returns
        -------
        bool
            True if the final estimator can fit_transform or is passthrough.
        """
        return (
            self._final_estimator == "passthrough"
            or hasattr(self._final_estimator, "transform")
            or hasattr(self._final_estimator, "fit_transform")
        )

    def _can_decision_function(self) -> bool:
        """Check if the final estimator implements decision_function.

        Returns
        -------
        bool
            True if the final estimator implements decision_function.
        """
        return hasattr(self._final_estimator, "decision_function")

    @available_if(_can_fit_transform)
    @_fit_context(
        # estimators in Pipeline.steps are not validated yet
        prefer_skip_nested_validation=False
    )
    def fit_transform(self, X: Any, y: Any = None, **params: Any) -> Any:
        """Fit the model and transform with the final estimator.

        Fits all the transformers one after the other and transform the
        data. Then uses `fit_transform` on transformed data with the final
        estimator.

        Parameters
        ----------
        X : iterable
            Training data. Must fulfill input requirements of first step of the
            pipeline.

        y : iterable, default=None
            Training targets. Must fulfill label requirements for all steps of
            the pipeline.

        **params : Any
            Parameters passed to the ``fit`` method of each step, where
            each parameter name is prefixed such that parameter ``p`` for step
            ``s`` has key ``s__p``.

        Returns
        -------
        Xt : ndarray of shape (n_samples, n_transformed_features)
            Transformed samples.
        """
        routed_params = self._check_method_params(method="fit_transform", props=params)
        iter_input, iter_label = self._fit(X, y, routed_params)
        last_step = self._final_estimator
        with print_elapsed_time("Pipeline", self._log_message(len(self.steps) - 1)):
            if last_step == "passthrough":
                pass
            elif is_empty(iter_input):
                logger.warning("All input rows were filtered out! Model is not fitted!")
            else:
                last_step_params = routed_params[
                    self._non_post_processing_steps()[-1][0]
                ]
                if hasattr(last_step, "fit_transform"):
                    iter_input = last_step.fit_transform(
                        iter_input, iter_label, **last_step_params["fit_transform"]
                    )
                elif hasattr(last_step, "transform") and hasattr(last_step, "fit"):
                    last_step.fit(iter_input, iter_label, **last_step_params["fit"])
                    iter_input = last_step.transform(
                        iter_input, **last_step_params["transform"]
                    )
                else:
                    raise TypeError(
                        f"fit_transform of the final estimator"
                        f" {last_step.__class__.__name__} {last_step_params} does not "
                        f"match fit_transform of Pipeline {self.__class__.__name__}"
                    )
            for _, post_element in self._post_processing_steps():
                iter_input = post_element.fit_transform(iter_input, iter_label)
        return iter_input

    @available_if(_final_estimator_has("predict"))
    def predict(self, X: Any, **params: Any) -> Any:
        """Transform the data, and apply `predict` with the final estimator.

        Call `transform` of each transformer in the pipeline. The transformed
        data are finally passed to the final estimator that calls `predict`
        method. Only valid if the final estimator implements `predict`.

        Parameters
        ----------
        X : iterable
            Data to predict on. Must fulfill input requirements of first step
            of the pipeline.

        **params : dict of string -> object
            Parameters to the ``predict`` called at the end of all
            transformations in the pipeline. Note that while this may be
            used to return uncertainties from some models with return_std
            or return_cov, uncertainties that are generated by the
            transformations in the pipeline are not propagated to the
            final estimator.

            .. versionadded:: 0.20

        Returns
        -------
        y_pred : ndarray
            Result of calling `predict` on the final estimator.
        """
        if _routing_enabled():
            routed_params = process_routing(self, "predict", **params)
        else:
            routed_params = process_routing(self, "predict", **{})

        iter_input = self._transform(X, routed_params)

        if self._final_estimator == "passthrough":
            pass
        elif is_empty(iter_input):
            iter_input = []
        elif hasattr(self._final_estimator, "predict"):
            if _routing_enabled():
                iter_input = self._final_estimator.predict(
                    iter_input,
                    **routed_params[self._non_post_processing_steps()[-1][0]].predict,
                )
            else:
                iter_input = self._final_estimator.predict(iter_input, **params)
        else:
            raise AssertionError(
                "Final estimator does not implement predict, hence this function should not be available."
            )
        for _, post_element in self._post_processing_steps():
            iter_input = post_element.transform(iter_input)
        return iter_input

    @available_if(_final_estimator_has("fit_predict"))
    @_fit_context(
        # estimators in Pipeline.steps are not validated yet
        prefer_skip_nested_validation=False
    )
    def fit_predict(self, X: Any, y: Any = None, **params: Any) -> Any:
        """Transform the data, and apply `fit_predict` with the final estimator.

        Call `fit_transform` of each transformer in the pipeline. The
        transformed data are finally passed to the final estimator that calls
        `fit_predict` method. Only valid if the final estimator implements
        `fit_predict`.

        Parameters
        ----------
        X : iterable
            Training data. Must fulfill input requirements of first step of
            the pipeline.

        y : iterable, default=None
            Training targets. Must fulfill label requirements for all steps
            of the pipeline.

        **params : dict of string -> object
            Parameters passed to the ``fit`` method of each step, where
            each parameter name is prefixed such that parameter ``p`` for step
            ``s`` has key ``s__p``.

        Returns
        -------
        y_pred : ndarray
            Result of calling `fit_predict` on the final estimator.
        """
        routed_params = self._check_method_params(method="fit_predict", props=params)
        iter_input, iter_label = self._fit(X, y, routed_params)

        params_last_step = routed_params[self._non_post_processing_steps()[-1][0]]
        with print_elapsed_time("Pipeline", self._log_message(len(self.steps) - 1)):
            if self._final_estimator == "passthrough":
                y_pred = iter_input
            elif is_empty(iter_input):
                logger.warning("All input rows were filtered out! Model is not fitted!")
                iter_input = []
                y_pred = []
            elif hasattr(self._final_estimator, "fit_predict"):
                y_pred = self._final_estimator.fit_predict(
                    iter_input, iter_label, **params_last_step.get("fit_predict", {})
                )
            else:
                raise AssertionError(
                    "Final estimator does not implement fit_predict, hence this function should not be available."
                )
            for _, post_element in self._post_processing_steps():
                y_pred = post_element.fit_transform(y_pred, iter_label)
        return y_pred

    @available_if(_final_estimator_has("predict_proba"))
    def predict_proba(self, X: Any, **params: Any) -> Any:
        """Transform the data, and apply `predict_proba` with the final estimator.

        Call `transform` of each transformer in the pipeline. The transformed
        data are finally passed to the final estimator that calls `predict_proba`
        method. Only valid if the final estimator implements `predict_proba`.

        Parameters
        ----------
        X : iterable
            Data to predict on. Must fulfill input requirements of first step
            of the pipeline.

        **params : dict of string -> object
            Parameters to the ``predict`` called at the end of all
            transformations in the pipeline. Note that while this may be
            used to return uncertainties from some models with return_std
            or return_cov, uncertainties that are generated by the
            transformations in the pipeline are not propagated to the
            final estimator.

            .. versionadded:: 0.20

        Returns
        -------
        y_pred : ndarray
            Result of calling `predict_proba` on the final estimator.
        """
        routed_params = process_routing(self, "predict_proba", **params)
        iter_input = self._transform(X, routed_params)

        if self._final_estimator == "passthrough":
            pass
        elif is_empty(iter_input):
            iter_input = []
        elif hasattr(self._final_estimator, "predict_proba"):
            if _routing_enabled():
                iter_input = self._final_estimator.predict_proba(
                    iter_input,
                    **routed_params[
                        self._non_post_processing_steps()[-1][0]
                    ].predict_proba,
                )
            else:
                iter_input = self._final_estimator.predict_proba(iter_input, **params)
        else:
            raise AssertionError(
                "Final estimator does not implement predict_proba, hence this function should not be available."
            )
        for _, post_element in self._post_processing_steps():
            iter_input = post_element.transform(iter_input)
        return iter_input

    def _can_transform(self) -> bool:
        """Check if the final estimator can transform or is passthrough.

        Returns
        -------
        bool
            True if the final estimator can transform or is passthrough.
        """
        return self._final_estimator == "passthrough" or hasattr(
            self._final_estimator, "transform"
        )

    @available_if(_can_transform)
    def transform(self, X: Any, **params: Any) -> Any:
        """Transform the data, and apply `transform` with the final estimator.

        Call `transform` of each transformer in the pipeline. The transformed
        data are finally passed to the final estimator that calls
        `transform` method. Only valid if the final estimator
        implements `transform`.

        This also works where final estimator is `None` in which case all prior
        transformations are applied.

        Parameters
        ----------
        X : iterable
            Data to transform. Must fulfill input requirements of first step
            of the pipeline.
        **params : Any
            Parameters to the ``transform`` method of each estimator.

        Returns
        -------
        Xt : ndarray of shape (n_samples, n_transformed_features)
            Transformed data.
        """
        routed_params = process_routing(self, "transform", **params)
        iter_input = X
        for _, name, transform in self._iter():
            if transform == "passthrough":
                continue
            if is_empty(iter_input):
                # This is done to prime the error filters
                if isinstance(transform, _MolPipeline):
                    _ = transform.transform(iter_input)
                iter_input = []
                break
            if hasattr(transform, "transform"):
                iter_input = transform.transform(
                    iter_input, **routed_params[name].transform
                )
            else:
                raise AssertionError(
                    "Non transformer ocurred in transformation step. This should have been caught in the validation step."
                )
        for _, post_element in self._post_processing_steps():
            iter_input = post_element.transform(iter_input, **params)
        return iter_input

    @available_if(_can_decision_function)
    def decision_function(self, X: Any, **params: Any) -> Any:
        """Transform the data, and apply `decision_function` with the final estimator.

        Parameters
        ----------
        X : iterable
            Data to transform. Must fulfill input requirements of first step
            of the pipeline.
        **params : Any
            Parameters to the ``decision_function`` method of the final estimator.

        Returns
        -------
        Any
            Result of calling `decision_function` on the final estimator.
        """
        if _routing_enabled():
            routed_params = process_routing(self, "decision_function", **params)
        else:
            routed_params = process_routing(self, "decision_function", **{})

        iter_input = self._transform(X, routed_params)
        if self._final_estimator == "passthrough":
            pass
        elif is_empty(iter_input):
            iter_input = []
        elif hasattr(self._final_estimator, "decision_function"):
            if _routing_enabled():
                iter_input = self._final_estimator.decision_function(
                    iter_input, **routed_params[self._final_estimator].predict
                )
            else:
                iter_input = self._final_estimator.decision_function(
                    iter_input, **params
                )
        else:
            raise AssertionError(
                "Final estimator does not implement `decision_function`, hence this function should not be available."
            )
        for _, post_element in self._post_processing_steps():
            iter_input = post_element.transform(iter_input)
        return iter_input

    @property
    def classes_(self) -> list[Any] | npt.NDArray[Any]:
        """Return the classes of the last element, which is not a PostPredictionTransformation."""
        check_last = [
            step
            for step in self.steps
            if not isinstance(step[1], PostPredictionTransformation)
        ]
        last_step = check_last[-1][1]
        if last_step == "passthrough":
            raise ValueError("Last step is passthrough.")
        if hasattr(last_step, "classes_"):
            return last_step.classes_
        raise ValueError("Last step has no classes_ attribute.")

    def __sklearn_tags__(self) -> Tags:
        """Return the sklearn tags.

        Note
        ----
        This method is copied from the original sklearn implementation.
        Changes are marked with a comment.

        Returns
        -------
        Tags
            The sklearn tags.
        """
        tags = super().__sklearn_tags__()

        if not self.steps:
            return tags

        try:
            if self.steps[0][1] is not None and self.steps[0][1] != "passthrough":
                tags.input_tags.pairwise = get_tags(
                    self.steps[0][1]
                ).input_tags.pairwise
            # WARNING: the sparse tag can be incorrect.
            # Some Pipelines accepting sparse data are wrongly tagged sparse=False.
            # For example Pipeline([PCA(), estimator]) accepts sparse data
            # even if the estimator doesn't as PCA outputs a dense array.
            tags.input_tags.sparse = all(
                get_tags(step).input_tags.sparse
                for name, step in self.steps
                if step != "passthrough"
            )
        except (ValueError, AttributeError, TypeError):
            # This happens when the `steps` is not a list of (name, estimator)
            # tuples and `fit` is not called yet to validate the steps.
            pass

        try:
            # Only the _final_estimator is changed from the original implementation is changed in the following 2 lines
            if (
                self._final_estimator is not None
                and self._final_estimator != "passthrough"
            ):
                last_step_tags = get_tags(self._final_estimator)
                tags.estimator_type = last_step_tags.estimator_type
                tags.target_tags.multi_output = last_step_tags.target_tags.multi_output
                tags.classifier_tags = deepcopy(last_step_tags.classifier_tags)
                tags.regressor_tags = deepcopy(last_step_tags.regressor_tags)
                tags.transformer_tags = deepcopy(last_step_tags.transformer_tags)
        except (ValueError, AttributeError, TypeError):
            # This happens when the `steps` is not a list of (name, estimator)
            # tuples and `fit` is not called yet to validate the steps.
            pass

        return tags

    def get_metadata_routing(self) -> MetadataRouter:
        """Get metadata routing of this object.

        Please check :ref:`User Guide <metadata_routing>` on how the routing
        mechanism works.

        Note
        ----
        This method is copied from the original sklearn implementation.
        Changes are marked with a comment.

        Returns
        -------
        MetadataRouter
            A :class:`~sklearn.utils.metadata_routing.MetadataRouter` encapsulating
            routing information.
        """
        router = MetadataRouter(owner=self.__class__.__name__)

        # first we add all steps except the last one
        for _, name, trans in self._iter(with_final=False, filter_passthrough=True):
            method_mapping = MethodMapping()
            # fit, fit_predict, and fit_transform call fit_transform if it
            # exists, or else fit and transform
            if hasattr(trans, "fit_transform"):
                (
                    method_mapping.add(caller="fit", callee="fit_transform")
                    .add(caller="fit_transform", callee="fit_transform")
                    .add(caller="fit_predict", callee="fit_transform")
                )
            else:
                (
                    method_mapping.add(caller="fit", callee="fit")
                    .add(caller="fit", callee="transform")
                    .add(caller="fit_transform", callee="fit")
                    .add(caller="fit_transform", callee="transform")
                    .add(caller="fit_predict", callee="fit")
                    .add(caller="fit_predict", callee="transform")
                )

            (
                method_mapping.add(caller="predict", callee="transform")
                .add(caller="predict", callee="transform")
                .add(caller="predict_proba", callee="transform")
                .add(caller="decision_function", callee="transform")
                .add(caller="predict_log_proba", callee="transform")
                .add(caller="transform", callee="transform")
                .add(caller="inverse_transform", callee="inverse_transform")
                .add(caller="score", callee="transform")
            )

            router.add(method_mapping=method_mapping, **{name: trans})

        # Only the _non_post_processing_steps is changed from the original implementation is changed in the following line
        final_name, final_est = self._non_post_processing_steps()[-1]
        if final_est is None or final_est == "passthrough":
            return router

        # then we add the last step
        method_mapping = MethodMapping()
        if hasattr(final_est, "fit_transform"):
            method_mapping.add(caller="fit_transform", callee="fit_transform")
        else:
            method_mapping.add(caller="fit", callee="fit").add(
                caller="fit", callee="transform"
            )
        (
            method_mapping.add(caller="fit", callee="fit")
            .add(caller="predict", callee="predict")
            .add(caller="fit_predict", callee="fit_predict")
            .add(caller="predict_proba", callee="predict_proba")
            .add(caller="decision_function", callee="decision_function")
            .add(caller="predict_log_proba", callee="predict_log_proba")
            .add(caller="transform", callee="transform")
            .add(caller="inverse_transform", callee="inverse_transform")
            .add(caller="score", callee="score")
        )

        router.add(method_mapping=method_mapping, **{final_name: final_est})
        return router



"""Tests for the MolToPathFingerprint pipeline element."""

from __future__ import annotations

import unittest
from typing import Any

import numpy as np

from molpipeline import Pipeline
from molpipeline.any2mol import SmilesToMol
from molpipeline.mol2any import Mol2PathFP

# pylint: disable=duplicate-code

test_smiles = [
    "c1ccccc1",
    "c1ccccc1C",
    "NCCOCCCC(=O)O",
]


class TestMol2PathFingerprint(unittest.TestCase):
    """Unittest for Mol2PathFP, which calculates the RDKit Path Fingerprint."""

    def test_can_be_constructed(self) -> None:
        """Test if the Mol2PathFP pipeline element can be constructed.

        Returns
        -------
        None
        """
        mol_fp = Mol2PathFP()
        mol_fp_copy = mol_fp.copy()
        self.assertTrue(mol_fp_copy is not mol_fp)
        for key, value in mol_fp.get_params().items():
            self.assertEqual(value, mol_fp_copy.get_params()[key])
        mol_fp_recreated = Mol2PathFP(**mol_fp.get_params())
        for key, value in mol_fp.get_params().items():
            self.assertEqual(value, mol_fp_recreated.get_params()[key])

    def test_output_types(self) -> None:
        """Test equality of different output_types."""

        smi2mol = SmilesToMol()
        sparse_path_fp = Mol2PathFP(n_bits=1024, return_as="sparse")
        dense_path_fp = Mol2PathFP(n_bits=1024, return_as="dense")
        explicit_bit_vect_path_fp = Mol2PathFP(
            n_bits=1024, return_as="explicit_bit_vect"
        )
        sparse_pipeline = Pipeline(
            [
                ("smi2mol", smi2mol),
                ("sparse_path_fp", sparse_path_fp),
            ],
        )
        dense_pipeline = Pipeline(
            [
                ("smi2mol", smi2mol),
                ("dense_path_fp", dense_path_fp),
            ],
        )
        explicit_bit_vect_pipeline = Pipeline(
            [
                ("smi2mol", smi2mol),
                ("explicit_bit_vect_path_fp", explicit_bit_vect_path_fp),
            ],
        )

        sparse_output = sparse_pipeline.fit_transform(test_smiles)
        dense_output = dense_pipeline.fit_transform(test_smiles)
        explicit_bit_vect_path_fp_output = explicit_bit_vect_pipeline.fit_transform(
            test_smiles
        )

        self.assertTrue(np.all(sparse_output.toarray() == dense_output))

        self.assertTrue(
            np.equal(
                dense_output,
                np.array(explicit_bit_vect_path_fp_output),
            ).all()
        )

    def test_counted_bits(self) -> None:
        """Test if the option counted bits works as expected.

        Returns
        -------
        None
        """
        mol_fp = Mol2PathFP(n_bits=1024, return_as="dense")
        smi2mol = SmilesToMol()
        pipeline = Pipeline(
            [
                ("smi2mol", smi2mol),
                ("mol_fp", mol_fp),
            ],
        )
        output_binary = pipeline.fit_transform(test_smiles)
        pipeline.set_params(mol_fp__counted=True)
        output_counted = pipeline.fit_transform(test_smiles)
        self.assertTrue(
            np.all(np.flatnonzero(output_counted) == np.flatnonzero(output_binary))
        )
        self.assertTrue(np.all(output_counted >= output_binary))
        self.assertTrue(np.any(output_counted > output_binary))

    def test_setter_getter(self) -> None:
        """Test if the setters and getters work as expected."""
        mol_fp = Mol2PathFP()
        params: dict[str, Any] = {
            "min_path": 10,
            "max_path": 12,
            "use_hs": False,
            "branched_paths": False,
            "use_bond_order": False,
            "count_simulation": True,
            "num_bits_per_feature": 4,
            "counted": True,
            "n_bits": 1024,
        }
        mol_fp.set_params(**params)
        self.assertEqual(mol_fp.get_params()["min_path"], 10)
        self.assertEqual(mol_fp.get_params()["max_path"], 12)
        self.assertEqual(mol_fp.get_params()["use_hs"], False)
        self.assertEqual(mol_fp.get_params()["branched_paths"], False)
        self.assertEqual(mol_fp.get_params()["use_bond_order"], False)
        self.assertEqual(mol_fp.get_params()["count_simulation"], True)
        self.assertEqual(mol_fp.get_params()["num_bits_per_feature"], 4)
        self.assertEqual(mol_fp.get_params()["counted"], True)
        self.assertEqual(mol_fp.get_params()["n_bits"], 1024)

    def test_setter_getter_error_handling(self) -> None:
        """Test if the setters and getters work as expected when errors are encountered."""

        mol_fp = Mol2PathFP()
        params: dict[str, Any] = {
            "min_path": 2,
            "n_bits": 1024,
            "return_as": "invalid-option",
        }
        self.assertRaises(ValueError, mol_fp.set_params, **params)

    def test_feature_names(self) -> None:
        """Test if the feature names are correct."""
        mol_fp = Mol2PathFP(n_bits=1024)
        feature_names = mol_fp.feature_names
        self.assertEqual(len(feature_names), 1024)
        # feature names should be unique
        self.assertEqual(len(feature_names), len(set(feature_names)))


if __name__ == "__main__":
    unittest.main()



"""Unittests for testing conversion of molecules to InChI and InChIKey."""

import unittest

from molpipeline.any2mol import SmilesToMol
from molpipeline.mol2any import MolToInchi, MolToInchiKey
from molpipeline.pipeline import Pipeline

# pylint: disable=duplicate-code  # test case molecules are allowed to be duplicated
SMILES_ANTIMONY = "[SbH6+3]"
SMILES_BENZENE = "c1ccccc1"
SMILES_CHLOROBENZENE = "Clc1ccccc1"
SMILES_CL_BR = "NC(Cl)(Br)C(=O)O"
SMILES_METAL_AU = "OC[C@H]1OC(S[Au])[C@H](O)[C@@H](O)[C@@H]1O"


class TestMol2Inchi(unittest.TestCase):
    """Test case for testing conversion of molecules to InChI and InChIKey."""

    def test_to_inchi(self) -> None:
        """Test if smiles converted correctly to inchi string.

        Returns
        -------
        None
        """

        input_smiles = ["CN(C)CCOC(C1=CC=CC=C1)C1=CC=CC=C1"]
        expected_inchis = [
            "InChI=1S/C17H21NO/c1-18(2)13-14-19-17(15-9-5-3-6-10-15)16-11-7-4-8-12-16/h3-12,17H,13-14H2,1-2H3"
        ]
        pipeline = Pipeline(
            [
                ("Smiles2Mol", SmilesToMol()),
                ("Mol2Inchi", MolToInchi()),
            ]
        )
        actual_inchis = pipeline.fit_transform(input_smiles)
        self.assertEqual(expected_inchis, actual_inchis)

    def test_to_inchikey(self) -> None:
        """Test if smiles is converted correctly to inchikey string.

        Returns
        -------
        None
        """

        input_smiles = ["CN(C)CCOC(C1=CC=CC=C1)C1=CC=CC=C1"]
        expected_inchikeys = ["ZZVUWRFHKOJYTH-UHFFFAOYSA-N"]

        pipeline = Pipeline(
            [
                ("Smiles2Mol", SmilesToMol()),
                ("Mol2Inchi", MolToInchiKey()),
            ],
        )
        actual_inchikeys = pipeline.fit_transform(input_smiles)
        self.assertEqual(expected_inchikeys, actual_inchikeys)


if __name__ == "__main__":
    unittest.main()



"""Logging helper functions."""

from __future__ import annotations

import timeit
from contextlib import contextmanager
from typing import Generator

from loguru import logger


def _message_with_time(source: str, message: str, time: float) -> str:
    """Create one line message for logging purposes.

    Adapted from sklearn's function to stay consistent with the logging style:
    https://github.com/scikit-learn/scikit-learn/blob/e16a6ddebd527e886fc22105710ee20ce255f9f0/sklearn/utils/_user_interface.py

    Parameters
    ----------
    source : str
        String indicating the source or the reference of the message.
    message : str
        Short message.
    time : float
        Time in seconds.

    Returns
    -------
    str
        Message with elapsed time.
    """
    start_message = f"[{source}] "

    # adapted from joblib.logger.short_format_time without the Windows -.1s
    # adjustment
    if time > 60:
        time_str = f"{(time / 60):4.1f}min"
    else:
        time_str = f" {time:5.1f}s"

    end_message = f" {message}, total={time_str}"
    dots_len = 70 - len(start_message) - len(end_message)
    return f"{start_message}{dots_len * '.'}{end_message}"


@contextmanager
def print_elapsed_time(
    source: str, message: str | None = None, use_logger: bool = False
) -> Generator[None, None, None]:
    """Log elapsed time to stdout when the context is exited.

    Adapted from sklearn's function to stay consistent with the logging style:
    https://github.com/scikit-learn/scikit-learn/blob/e16a6ddebd527e886fc22105710ee20ce255f9f0/sklearn/utils/_user_interface.py

    Parameters
    ----------
    source : str
        String indicating the source or the reference of the message.
    message : str, default=None
        Short message. If None, nothing will be printed.
    use_logger : bool, default=False
        If True, the message will be logged using the logger.

    Returns
    -------
    context_manager
        Prints elapsed time upon exit if verbose.
    """
    if message is None:
        yield
    else:
        start = timeit.default_timer()
        yield
        message_to_print = _message_with_time(
            source, message, timeit.default_timer() - start
        )

        if use_logger:
            logger.info(message_to_print)
        else:
            print(message_to_print)



"""Test construction of concatenated fingerprints."""

from __future__ import annotations

import itertools
import unittest
from typing import Any, Literal, get_args

import numpy as np
from rdkit import Chem
from sklearn.preprocessing import StandardScaler

from molpipeline import Pipeline
from molpipeline.abstract_pipeline_elements.core import MolToAnyPipelineElement
from molpipeline.any2mol import SmilesToMol
from molpipeline.mol2any import (
    Mol2PathFP,
    MolToConcatenatedVector,
    MolToMACCSFP,
    MolToMorganFP,
    MolToNetCharge,
    MolToRDKitPhysChem,
)
from tests.utils.fingerprints import fingerprints_to_numpy
from tests.utils.logging import capture_logs


class TestConcatenatedFingerprint(unittest.TestCase):
    """Unittest for MolToConcatenatedVector, which calculates concatenated fingerprints."""

    def test_generation(self) -> None:
        """Test if the feature concatenation works as expected."""
        fingerprint_morgan_output_types: tuple[Any, ...] = get_args(
            Literal[
                "sparse",
                "dense",
                "explicit_bit_vect",
            ]
        )

        smiles = [
            "CC",
            "CCC",
            "CCCO",
            "CCNCO",
            "C(C)CCO",
            "CCO",
            "CCCN",
            "CCCC",
            "CCOC",
            "COO",
        ]

        for fp_output_type in fingerprint_morgan_output_types:

            concat_vector_element = MolToConcatenatedVector(
                [
                    (
                        "RDKitPhysChem",
                        MolToRDKitPhysChem(standardizer=StandardScaler()),
                    ),
                    (
                        "MorganFP",
                        MolToMorganFP(return_as=fp_output_type),
                    ),
                ]
            )
            pipeline = Pipeline(
                [
                    ("smi2mol", SmilesToMol()),
                    ("concat_vector_element", concat_vector_element),
                ]
            )

            output = pipeline.fit_transform(smiles)
            output2 = pipeline.transform(smiles)

            mol_list = [Chem.MolFromSmiles(smi) for smi in smiles]
            output3 = np.hstack(
                [
                    concat_vector_element.element_list[0][1].transform(mol_list),
                    fingerprints_to_numpy(
                        concat_vector_element.element_list[1][1].transform(mol_list)
                    ),
                ]
            )
            pyschem_component: MolToRDKitPhysChem
            pyschem_component = concat_vector_element.element_list[0][1]  # type: ignore
            morgan_component: MolToMorganFP
            morgan_component = concat_vector_element.element_list[1][1]  # type: ignore
            expected_shape = (
                len(smiles),
                (pyschem_component.n_features + morgan_component.n_bits),
            )
            self.assertEqual(output.shape, expected_shape)
            self.assertTrue(np.allclose(output, output2))
            self.assertTrue(np.allclose(output, output3))

    def test_empty_element_list(self) -> None:
        """Test if an empty element list raises an error."""
        # test constructor
        with self.assertRaises(ValueError):
            MolToConcatenatedVector([])

        # test setter
        concat_elem = MolToConcatenatedVector(
            [
                (
                    "RDKitPhysChem",
                    MolToRDKitPhysChem(),
                )
            ]
        )
        with self.assertRaises(ValueError):
            concat_elem.set_params(element_list=[])

    def test_n_features(self) -> None:
        """Test getting the number of features in the concatenated vector."""

        physchem_elem = (
            "RDKitPhysChem",
            MolToRDKitPhysChem(),
        )
        morgan_elem = (
            "MorganFP",
            MolToMorganFP(n_bits=16),
        )
        net_charge_elem = ("NetCharge", MolToNetCharge())

        self.assertEqual(
            MolToConcatenatedVector([physchem_elem]).n_features,
            physchem_elem[1].n_features,
        )
        self.assertEqual(
            MolToConcatenatedVector([morgan_elem]).n_features,
            16,
        )
        self.assertEqual(
            MolToConcatenatedVector([net_charge_elem]).n_features,
            net_charge_elem[1].n_features,
        )
        self.assertEqual(
            MolToConcatenatedVector([physchem_elem, morgan_elem]).n_features,
            physchem_elem[1].n_features + 16,
        )
        self.assertEqual(
            MolToConcatenatedVector([net_charge_elem, morgan_elem]).n_features,
            net_charge_elem[1].n_features + 16,
        )
        self.assertEqual(
            MolToConcatenatedVector(
                [net_charge_elem, morgan_elem, physchem_elem]
            ).n_features,
            net_charge_elem[1].n_features + 16 + physchem_elem[1].n_features,
        )

    def test_features_names(self) -> None:  # pylint: disable-msg=too-many-locals
        """Test getting the names of features in the concatenated vector."""

        physchem_elem = (
            "RDKitPhysChem",
            MolToRDKitPhysChem(),
        )
        net_charge_elem = ("NetCharge", MolToNetCharge())
        morgan_elem = (
            "MorganFP",
            MolToMorganFP(n_bits=17),
        )
        path_elem = (
            "PathFP",
            Mol2PathFP(n_bits=15),
        )
        maccs_elem = (
            "MACCSFP",
            MolToMACCSFP(),
        )

        elements = [
            physchem_elem,
            net_charge_elem,
            morgan_elem,
            path_elem,
            maccs_elem,
        ]

        for use_feature_names_prefix in [False, True]:
            # test all subsets are compatible
            powerset = itertools.chain.from_iterable(
                itertools.combinations(elements, r) for r in range(len(elements) + 1)
            )
            # skip empty subset
            next(powerset)

            for elements_subset in powerset:
                conc_elem = MolToConcatenatedVector(
                    list(elements_subset),
                    use_feature_names_prefix=use_feature_names_prefix,
                )
                feature_names = conc_elem.feature_names

                if use_feature_names_prefix:
                    # test feature names are unique if prefix is used or only one element is used
                    self.assertEqual(
                        len(feature_names),
                        len(set(feature_names)),
                    )

                # test a feature names and n_features are consistent
                self.assertEqual(
                    len(feature_names),
                    conc_elem.n_features,
                )

                seen_names = 0
                for elem_name, elem in elements_subset:
                    self.assertTrue(hasattr(elem, "feature_names"))
                    elem_feature_names = elem.feature_names  # type: ignore[attr-defined]
                    elem_n_features = len(elem_feature_names)
                    relevant_names = feature_names[
                        seen_names : seen_names + elem_n_features
                    ]

                    if use_feature_names_prefix:
                        # feature_names should be prefixed with element name
                        prefixes, feat_names = map(
                            list, zip(*[name.split("__") for name in relevant_names])
                        )
                        # test feature names are the same
                        self.assertListEqual(elem_feature_names, feat_names)
                        # test prefixes are the same as element names
                        self.assertTrue(all(prefix == elem_name for prefix in prefixes))
                    else:
                        # feature_names should be the same as element feature names
                        self.assertListEqual(elem_feature_names, relevant_names)

                    seen_names += elem_n_features

    def test_logging_feature_names_uniqueness(self) -> None:
        """Test that a warning is logged when feature names are not unique."""
        elements: list[tuple[str, MolToAnyPipelineElement]] = [
            (
                "MorganFP",
                MolToMorganFP(n_bits=17),
            ),
            (
                "MorganFP_with_feats",
                MolToMorganFP(n_bits=16, use_features=True),
            ),
        ]

        # First test is with no prefix
        use_feature_names_prefix = False
        with capture_logs() as output:
            conc_elem = MolToConcatenatedVector(
                elements,
                use_feature_names_prefix=use_feature_names_prefix,
            )
            feature_names = conc_elem.feature_names

        # test log message
        self.assertEqual(len(output), 1)
        message = output[0]
        self.assertIn(
            "Feature names in MolToConcatenatedVector are not unique", message
        )
        self.assertEqual(message.record["level"].name, "WARNING")

        # test feature names are NOT unique
        self.assertNotEqual(len(feature_names), len(set(feature_names)))

        # Second test is with prefix
        use_feature_names_prefix = True
        with capture_logs() as output:
            conc_elem = MolToConcatenatedVector(
                elements,
                use_feature_names_prefix=use_feature_names_prefix,
            )
            feature_names = conc_elem.feature_names

        # test log message
        self.assertEqual(len(output), 0)

        # test feature names are unique
        self.assertEqual(len(feature_names), len(set(feature_names)))

    def test_getter_setter(self) -> None:
        """Test getter and setter methods."""
        elements: list[tuple[str, MolToAnyPipelineElement]] = [
            (
                "MorganFP",
                MolToMorganFP(n_bits=17),
            ),
            (
                "MorganFP_with_feats",
                MolToMorganFP(n_bits=16, use_features=True),
            ),
        ]
        concat_elem = MolToConcatenatedVector(
            elements,
            use_feature_names_prefix=True,
        )
        self.assertEqual(len(concat_elem.get_params()["element_list"]), 2)
        self.assertEqual(concat_elem.get_params()["use_feature_names_prefix"], True)
        # test that there are no duplicates in feature names
        self.assertEqual(
            len(concat_elem.feature_names), len(set(concat_elem.feature_names))
        )
        params: dict[str, Any] = {
            "use_feature_names_prefix": False,
        }
        concat_elem.set_params(**params)
        self.assertEqual(concat_elem.get_params()["use_feature_names_prefix"], False)
        # test that there are duplicates in feature names
        self.assertNotEqual(
            len(concat_elem.feature_names), len(set(concat_elem.feature_names))
        )


if __name__ == "__main__":
    unittest.main()



"""LeaderPicker-based clustering estimator."""

from __future__ import annotations

from itertools import compress
from numbers import Real

import numpy as np
import numpy.typing as npt
from rdkit import DataStructs
from rdkit.DataStructs import ExplicitBitVect
from rdkit.SimDivFilters import rdSimDivPickers
from sklearn.base import BaseEstimator, ClusterMixin, _fit_context
from sklearn.utils._param_validation import Interval

try:
    from typing import Any, Self, Sequence
except ImportError:
    from typing_extensions import Self


class LeaderPickerClustering(ClusterMixin, BaseEstimator):
    """LeaderPicker clustering estimator (a sphere exclusion clustering algorithm)."""

    # we use sklearn's input validation to check constraints
    _parameter_constraints: dict[str, Any] = {
        "distance_threshold": [Interval(Real, 0, 1.0, closed="left")],
    }

    def __init__(
        self,
        distance_threshold: float,
    ) -> None:
        """Initialize LeaderPicker clustering estimator.

        Parameters
        ----------
        distance_threshold : float
            Minimum distance between cluster centroids.
        """
        self.distance_threshold: float = distance_threshold
        self.n_clusters_: int | None = None
        self.labels_: npt.NDArray[np.int32] | None = None
        # centroid indices
        self.centroids_: npt.NDArray[np.int32] | None = None

    # pylint: disable=C0103,W0613
    @_fit_context(prefer_skip_nested_validation=True)
    def fit(
        self,
        X: list[ExplicitBitVect],
        y: npt.NDArray[np.float64] | None = None,
    ) -> Self:
        """Fit leader picker clustering estimator.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            Feature matrix.
        y : Ignored
            Not used, present for API consistency by convention.

        Returns
        -------
        Self
            Fitted estimator.
        """
        return self._fit(X)

    @staticmethod
    def _assign_points_to_clusters_based_on_centroid(
        picks: Sequence[int], fps: Sequence[ExplicitBitVect]
    ) -> tuple[int, npt.NDArray[np.int32]]:
        """Assign points to clusters based on centroid.

        Based on https://rdkit.blogspot.com/2020/11/sphere-exclusion-clustering-with-rdkit.html

        Parameters
        ----------
        picks : Sequence[int]
            Indices of selected cluster centroids to which the remaining data will be assigned.
        fps : Sequence[ExplicitBitVect]
            Fingerprints of the whole data sets.

        Returns
        -------
        tuple[int, np.ndarray[int]]
            Number of clusters and cluster labels.
        """
        labels: npt.NDArray[np.int32] = np.full(len(fps), -1, dtype=np.int32)
        max_similarities = np.full(len(fps), -np.inf, dtype=np.float64)

        for i, pick_idx in enumerate(picks):
            similarities = DataStructs.BulkTanimotoSimilarity(fps[pick_idx], fps)
            max_mask = similarities > max_similarities
            labels[max_mask] = i
            max_similarities[max_mask] = list(compress(similarities, max_mask))

        return np.unique(labels).shape[0], labels

    # pylint: disable=C0103,W0613
    def _fit(self, X: list[ExplicitBitVect]) -> Self:
        """Fit leader picker clustering estimator.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            Feature matrix.

        Returns
        -------
        Self
            Fitted estimator.
        """
        lp = rdSimDivPickers.LeaderPicker()

        # Select centroids. This part is in C++ and fast
        picks = lp.LazyBitVectorPick(
            objects=X,
            poolSize=len(X),
            threshold=self.distance_threshold,
            numThreads=1,  # according to rdkit docu this parameter is not used
            # seed=self.random_state if self.random_state is not None else -1,
        )

        # Assign points to clusters based on centroid
        (
            self.n_clusters_,
            self.labels_,
        ) = self._assign_points_to_clusters_based_on_centroid(picks, X)

        self.centroids_ = np.array(picks)
        return self

    def fit_predict(
        self,
        X: list[ExplicitBitVect],  # pylint: disable=C0103
        y: npt.NDArray[np.float64] | None = None,
        **kwargs: Any,
    ) -> npt.NDArray[np.int32]:
        """Fit and predict leader picker clustering estimator.

        Parameters
        ----------
        X: npt.NDArray[np.float64] | csr_matrix
            Feature matrix of shape  (n_samples, n_features).
        y: Ignored
            Not used, present for API consistency by convention.
        kwargs: Any
            Additional keyword arguments.

        Returns
        -------
        np.ndarray[int]
            Cluster labels.
        """
        # pylint: disable=W0246
        return super().fit_predict(X, y, **kwargs)



"""Utility functions for explainability."""

from __future__ import annotations

from collections import defaultdict
from typing import Sequence

import numpy as np
import numpy.typing as npt

from molpipeline.abstract_pipeline_elements.core import RDKitMol
from molpipeline.mol2any import MolToMorganFP
from molpipeline.utils.substructure_handling import AtomEnvironment


def assign_prediction_importance(
    bit_dict: dict[int, Sequence[AtomEnvironment]], weights: npt.NDArray[np.float64]
) -> dict[int, float]:
    """Assign the prediction importance.

    Originally from Christian W. Feldmann
    https://github.com/c-feldmann/compchemkit/blob/64e5543e2b8f72e93711186b2e0b42366820fb52/compchemkit/molecular_heatmaps.py#L28

    Parameters
    ----------
    bit_dict : dict[int, Sequence[AtomEnvironment]]
        The bit dictionary.
    weights : npt.NDArray[np.float64]
        The weights.

    Returns
    -------
    dict[int, float]
        The atom contribution.
    """
    atom_contribution: dict[int, float] = defaultdict(lambda: 0)
    for bit, atom_env_list in bit_dict.items():  # type: int, Sequence[AtomEnvironment]
        n_machtes = len(atom_env_list)
        for atom_set in atom_env_list:
            for atom in atom_set.environment_atoms:
                atom_contribution[atom] += weights[bit] / (
                    len(atom_set.environment_atoms) * n_machtes
                )
    if not np.isclose(sum(weights), sum(atom_contribution.values())).all():
        raise AssertionError(
            f"Weights and atom contributions don't sum to the same value:"
            f" {weights.sum()} != {sum(atom_contribution.values())}"
        )
    return atom_contribution


def fingerprint_shap_to_atomweights(
    mol: RDKitMol, fingerprint_element: MolToMorganFP, shap_mat: npt.NDArray[np.float64]
) -> list[float]:
    """Convert SHAP values to atom weights.

    Originally from Christian W. Feldmann
    https://github.com/c-feldmann/compchemkit/blob/64e5543e2b8f72e93711186b2e0b42366820fb52/compchemkit/molecular_heatmaps.py#L15

    Parameters
    ----------
    mol : RDKitMol
        The molecule.
    fingerprint_element : MolToMorganFP
        The fingerprint element.
    shap_mat : npt.NDArray[np.float64]
        The SHAP values.

    Returns
    -------
    list[float]
        The atom weights.
    """
    bit_atom_env_dict: dict[int, Sequence[AtomEnvironment]]
    bit_atom_env_dict = dict(
        fingerprint_element.bit2atom_mapping(mol)
    )  # MyPy invariants make me do this.
    atom_weight_dict = assign_prediction_importance(bit_atom_env_dict, shap_mat)
    atom_weight_list = [
        atom_weight_dict[a_idx] if a_idx in atom_weight_dict else 0
        for a_idx in range(mol.GetNumAtoms())
    ]
    return atom_weight_list



"""Test Chemprop neural fingerprint."""

import logging
import unittest

from sklearn.base import clone

from molpipeline.estimators.chemprop.neural_fingerprint import ChempropNeuralFP
from molpipeline.utils.json_operations import recursive_from_json, recursive_to_json
from test_extras.test_chemprop.chemprop_test_utils.compare_models import compare_params
from test_extras.test_chemprop.chemprop_test_utils.default_models import (
    get_neural_fp_encoder,
)

logging.getLogger("lightning.pytorch.utilities.rank_zero").setLevel(logging.WARNING)


class TestChempropNeuralFingerprint(unittest.TestCase):
    """Test the Chemprop model."""

    def test_clone(self) -> None:
        """Test the clone method."""
        chemprop_fp_encoder = get_neural_fp_encoder()
        cloned_encoder = clone(chemprop_fp_encoder)
        self.assertIsInstance(cloned_encoder, ChempropNeuralFP)
        compare_params(self, chemprop_fp_encoder, cloned_encoder)

    def test_json_serialization(self) -> None:
        """Test the to_json and from_json methods."""
        chemprop_fp_encoder = get_neural_fp_encoder()
        chemprop_json = recursive_to_json(chemprop_fp_encoder)
        chemprop_encoder_copy = recursive_from_json(chemprop_json)
        compare_params(self, chemprop_fp_encoder, chemprop_encoder_copy)

    def test_output_type(self) -> None:
        """Test the output type."""
        chemprop_fp_encoder = get_neural_fp_encoder()
        self.assertEqual(chemprop_fp_encoder.output_type, "float")

    def test_init_with_kwargs(self) -> None:
        """Test the __init__ method with kwargs."""
        init_kwargs = {"model__message_passing__depth": 4}
        chemprop_fp_encoder = get_neural_fp_encoder(init_kwargs=init_kwargs)
        deep_params = chemprop_fp_encoder.get_params(deep=True)
        self.assertEqual(deep_params["model__message_passing__depth"], 4)



"""Unittests for testing conversion of binary string to RDKit molecules."""

import unittest

from rdkit import Chem, rdBase

from molpipeline.any2mol import BinaryToMol

# pylint: disable=duplicate-code  # test case molecules are allowed to be duplicated
SMILES_ANTIMONY = "[SbH6+3]"
SMILES_BENZENE = "c1ccccc1"
SMILES_CHLOROBENZENE = "Clc1ccccc1"
SMILES_CL_BR = "NC(Cl)(Br)C(=O)O"
SMILES_METAL_AU = "OC[C@H]1OC(S[Au])[C@H](O)[C@@H](O)[C@@H]1O"

# RDKit mols
MOL_ANTIMONY = Chem.MolFromSmiles(SMILES_ANTIMONY)
MOL_BENZENE = Chem.MolFromSmiles(SMILES_BENZENE)
MOL_CHLOROBENZENE = Chem.MolFromSmiles(SMILES_CHLOROBENZENE)
MOL_CL_BR = Chem.MolFromSmiles(SMILES_CL_BR)
MOL_METAL_AU = Chem.MolFromSmiles(SMILES_METAL_AU)


class TestBin2Mol(unittest.TestCase):
    """Test case for testing conversion of binary string to molecules."""

    def test_bin2mol(self) -> None:
        """Test molecules can be read from binary string."""
        test_mols = [
            MOL_ANTIMONY,
            MOL_BENZENE,
            MOL_CHLOROBENZENE,
            MOL_CL_BR,
            MOL_METAL_AU,
        ]
        for mol in test_mols:
            bin2mol = BinaryToMol()
            transformed_mol = bin2mol.pretransform_single(mol.ToBinary())
            log_block = rdBase.BlockLogs()
            self.assertEqual(
                Chem.MolToInchi(transformed_mol), Chem.MolToInchi(transformed_mol)
            )
            del log_block



"""Implementation of MACCS key fingerprint."""

import numpy as np
from numpy import typing as npt
from rdkit.Chem import MACCSkeys
from rdkit.DataStructs import ExplicitBitVect

from molpipeline.abstract_pipeline_elements.mol2any.mol2bitvector import (
    MolToFingerprintPipelineElement,
)
from molpipeline.utils.molpipeline_types import RDKitMol


class MolToMACCSFP(MolToFingerprintPipelineElement):
    """MACCS key fingerprint.

    The MACCS keys are a set of 166 keys that encode the presence or absence of
    particular substructures in a molecule. The MACCS keys are a subset of the
    PubChem substructure keys.

    """

    _n_bits = 167  # MACCS keys have 166 bits + 1 bit for an all-zero vector (bit 0)
    _feature_names = [f"maccs_{i}" for i in range(_n_bits)]

    def pretransform_single(
        self, value: RDKitMol
    ) -> dict[int, int] | npt.NDArray[np.int_] | ExplicitBitVect:
        """Transform a single molecule to MACCS key fingerprint.

        Parameters
        ----------
        value : RDKitMol
            RDKit molecule.

        Returns
        -------
        dict[int, int] | npt.NDArray[np.int_] | ExplicitBitVect
            MACCS key fingerprint.

        """
        fingerprint = MACCSkeys.GenMACCSKeys(value)  # type: ignore[attr-defined]
        if self._return_as == "explicit_bit_vect":
            return fingerprint
        if self._return_as == "dense":
            return np.array(fingerprint)
        if self._return_as == "sparse":
            return {idx: 1 for idx in fingerprint.GetOnBits()}
        raise ValueError(f"Unknown return_as value: {self._return_as}")



"""Definition of types used in molpipeline."""

from __future__ import annotations

from numbers import Number
from typing import (
    Any,
    Literal,
    Optional,
    Protocol,
    Sequence,
    Tuple,
    TypeAlias,
    TypeVar,
    Union,
)

try:
    from typing import Self  # type: ignore[attr-defined]
except ImportError:
    from typing_extensions import Self

import numpy as np
import numpy.typing as npt

from molpipeline.abstract_pipeline_elements.core import (
    ABCPipelineElement,
    OptionalMol,
    RDKitMol,
)

__all__ = [
    "AnyElement",
    "AnyNumpyElement",
    "AnyPredictor",
    "AnySklearnEstimator",
    "AnyTransformer",
    "AnySklearnEstimator",
    "AnyStep",
    "Number",
    "OptionalMol",
    "RDKitMol",
]
# One liner type definitions

AnyNumpyElement = TypeVar("AnyNumpyElement", bound=np.generic)

_T = TypeVar("_T")
_NT = TypeVar("_NT", bound=np.generic)
TypeFixedVarSeq = TypeVar("TypeFixedVarSeq", bound=Sequence[_T] | npt.NDArray[_NT])  # type: ignore
AnyVarSeq = TypeVar("AnyVarSeq", bound=Sequence[Any] | npt.NDArray[Any])

FloatCountRange: TypeAlias = tuple[Optional[float], Optional[float]]
IntCountRange: TypeAlias = tuple[Optional[int], Optional[int]]

# IntOrIntCountRange for Typing of count ranges
# - a single int for an exact value match
# - a range given as a tuple with a lower and upper bound
#   - both limits are optional
IntOrIntCountRange: TypeAlias = Union[int, IntCountRange]


class AnySklearnEstimator(Protocol):
    """Protocol for sklearn estimators."""

    def get_params(self, deep: bool = True) -> dict[str, Any]:
        """Get parameters for this estimator.

        Parameters
        ----------
        deep: bool
            If True, will return the parameters for this estimator.

        Returns
        -------
        dict[str, Any]
            Parameter names mapped to their values.
        """

    def set_params(self, **params: Any) -> Self:
        """Set the parameters of this estimator.

        Parameters
        ----------
        params: Any
            Estimator parameters.

        Returns
        -------
        Self
            Estimator with updated parameters.
        """

    def fit(
        self,
        X: npt.NDArray[Any],  # pylint: disable=invalid-name
        y: Optional[npt.NDArray[Any]],  # pylint: disable=invalid-name
        **fit_params: Any,
    ) -> Self:
        """Fit the model with X.

        Parameters
        ----------
        X: npt.NDArray[Any]
            Model input.
        y: Optional[npt.NDArray[Any]]
            Target values.
        fit_params: Any
            Additional parameters for fitting.


        Returns
        -------
        Self
            Fitted estimator.
        """


class AnyPredictor(AnySklearnEstimator, Protocol):
    """Protocol for predictors."""

    def fit_predict(
        self,
        X: npt.NDArray[Any],  # pylint: disable=invalid-name
        y: Optional[npt.NDArray[Any]],  # pylint: disable=invalid-name
        **fit_params: Any,
    ) -> npt.NDArray[Any]:
        """Fit the model with X and return predictions.

        Parameters
        ----------
        X: npt.NDArray[Any]
            Model input.
        y: Optional[npt.NDArray[Any]]
            Target values.
        fit_params: Any
            Additional parameters for fitting.

        Returns
        -------
        npt.NDArray[Any]
            Predictions.
        """


class AnyTransformer(AnySklearnEstimator, Protocol):
    """Protocol for transformers."""

    def fit_transform(
        self,
        X: npt.NDArray[Any],  # pylint: disable=invalid-name
        y: Optional[npt.NDArray[Any]],  # pylint: disable=invalid-name
        **fit_params: Any,
    ) -> npt.NDArray[Any]:
        """Fit the model with X and return the transformed array.

        Parameters
        ----------
        X: npt.NDArray[Any]
            Model input.
        y: Optional[npt.NDArray[Any]]
            Target values.
        fit_params: Any
            Additional parameters for fitting.


        Returns
        -------
        npt.NDArray[Any]
            Transformed array.
        """

    def transform(
        self,
        X: npt.NDArray[Any],  # pylint: disable=invalid-name
        **params: Any,
    ) -> npt.NDArray[Any]:
        """Transform and return X according to object protocol.

        Parameters
        ----------
        X: npt.NDArray[Any]
            Model input.
        params: Any
            Additional parameters for transforming.

        Returns
        -------
        npt.NDArray[Any]
            Transformed array.
        """


AnyElement = Union[
    AnyTransformer, AnyPredictor, ABCPipelineElement, Literal["passthrough"]
]
AnyStep = Tuple[str, AnyElement]



"""Visualization functions for the explainability module.

Much of the visualization code in this file originates from projects of Christian W. Feldmann:
    https://github.com/c-feldmann/rdkit_heatmaps
    https://github.com/c-feldmann/compchemkit
"""

from __future__ import annotations

from typing import Sequence

import numpy as np
import numpy.typing as npt
from matplotlib import colors
from matplotlib import pyplot as plt
from matplotlib.colors import Colormap
from matplotlib.figure import Figure
from PIL import Image
from rdkit import Chem
from rdkit.Chem import Draw
from rdkit.Chem.Draw import rdMolDraw2D

from molpipeline.abstract_pipeline_elements.core import RDKitMol
from molpipeline.experimental.explainability.explanation import (
    SHAPFeatureAndAtomExplanation,
)
from molpipeline.experimental.explainability.visualization.gauss import GaussFunctor2D
from molpipeline.experimental.explainability.visualization.heatmaps import (
    ValueGrid,
    color_canvas,
    get_color_normalizer_from_data,
)
from molpipeline.experimental.explainability.visualization.utils import (
    RGBAtuple,
    calc_present_and_absent_shap_contributions,
    get_atom_coords_of_bond,
    get_color_map_from_input,
    get_mol_lims,
    pad,
    plt_to_pil,
    to_png,
)


def _make_grid_from_mol(
    mol: Chem.Mol,
    grid_resolution: Sequence[int],
    padding: Sequence[float],
) -> ValueGrid:
    """Create a grid for the molecule.

    Parameters
    ----------
    mol: Chem.Mol
        RDKit molecule object.
    grid_resolution: Sequence[int]
        Resolution of the grid.
    padding: Sequence[float]
        Padding of the grid.

    Returns
    -------
    ValueGrid
        ValueGrid object.
    """
    xl: list[float]
    yl: list[float]
    xl, yl = [list(lim) for lim in get_mol_lims(mol)]  # Limit of molecule

    # Extent of the canvas is approximated by size of molecule scaled by ratio of canvas height and width.
    # Would be nice if this was directly accessible...
    mol_height = yl[1] - yl[0]
    mol_width = xl[1] - xl[0]

    height_to_width_ratio_mol = mol_height / (1e-16 + mol_width)
    # the grids height / weight is the canvas height / width
    height_to_width_ratio_canvas = grid_resolution[1] / grid_resolution[0]

    if height_to_width_ratio_mol < height_to_width_ratio_canvas:
        mol_height_new = height_to_width_ratio_canvas * mol_width
        yl[0] -= (mol_height_new - mol_height) / 2
        yl[1] += (mol_height_new - mol_height) / 2
    else:
        mol_width_new = grid_resolution[0] / grid_resolution[1] * mol_height
        xl[0] -= (mol_width_new - mol_width) / 2
        xl[1] += (mol_width_new - mol_width) / 2

    xl = list(pad(xl, padding[0]))  # Increasing size of x-axis
    yl = list(pad(yl, padding[1]))  # Increasing size of y-axis
    v_map = ValueGrid(xl, yl, grid_resolution[0], grid_resolution[1])
    return v_map


def _add_gaussians_for_atoms(
    mol: Chem.Mol,
    conf: Chem.Conformer,
    v_map: ValueGrid,
    atom_weights: npt.NDArray[np.float64],
    atom_width: float,
) -> ValueGrid:
    """Add Gauss-functions centered at atoms to the grid.

    Parameters
    ----------
    mol: Chem.Mol
        RDKit molecule object.
    conf: Chem.Conformer
        Conformation of the molecule.
    v_map: ValueGrid
        ValueGrid object to which the functions are added.
    atom_weights: npt.NDArray[np.float64]
        Array of weights for atoms.
    atom_width: float
        Width of the displayed atom weights.

    Returns
    -------
    ValueGrid
        ValueGrid object with added functions.
    """
    for i in range(mol.GetNumAtoms()):
        if atom_weights[i] == 0:
            continue
        pos = conf.GetAtomPosition(i)
        coords = np.array([pos.x, pos.y])
        func = GaussFunctor2D(
            center=coords,
            std1=atom_width,
            std2=atom_width,
            scale=atom_weights[i],
            rotation=0,
        )
        v_map.add_function(func)
    return v_map


def _add_gaussians_for_bonds(
    mol: Chem.Mol,
    conf: Chem.Conformer,
    v_map: ValueGrid,
    bond_weights: npt.NDArray[np.float64],
    bond_width: float,
    bond_length: float,
) -> ValueGrid:  # pylint: disable=too-many-locals
    """Add Gauss-functions centered at bonds to the grid.

    Parameters
    ----------
    mol: Chem.Mol
        RDKit molecule object.
    conf: Chem.Conformer
        Conformation of the molecule.
    v_map: ValueGrid
        ValueGrid object to which the functions are added.
    bond_weights: npt.NDArray[np.float64]
        Array of weights for bonds.
    bond_width: float
        Width of the displayed bond weights (perpendicular to bond-axis).
    bond_length: float
        Length of the displayed bond weights (along the bond-axis).

    Returns
    -------
    ValueGrid
        ValueGrid object with added functions.
    """
    # Adding Gauss-functions centered at bonds (position between the two bonded-atoms)
    for i, bond in enumerate(mol.GetBonds()):
        if bond_weights[i] == 0:
            continue

        a1_coords, a2_coords = get_atom_coords_of_bond(bond, conf)

        diff = a2_coords - a1_coords
        angle = np.arctan2(diff[0], diff[1])

        bond_center = (a1_coords + a2_coords) / 2

        func = GaussFunctor2D(
            center=bond_center,  # type: ignore
            std1=bond_width,
            std2=bond_length,
            scale=bond_weights[i],
            rotation=angle,
        )
        v_map.add_function(func)
    return v_map


def make_sum_of_gaussians_grid(
    mol: Chem.Mol,
    grid_resolution: Sequence[int],
    padding: Sequence[float],
    atom_weights: Sequence[float] | npt.NDArray[np.float64] | None = None,
    bond_weights: Sequence[float] | npt.NDArray[np.float64] | None = None,
    atom_width: float = 0.3,
    bond_width: float = 0.25,
    bond_length: float = 0.5,
) -> rdMolDraw2D:
    """Map weights of atoms and bonds to the drawing of a RDKit molecular depiction.

    For each atom and bond of depicted molecule a Gauss-function, centered at the respective object, is created and
    scaled by the corresponding weight. Gauss-functions of atoms are circular, while Gauss-functions of bonds can be
    distorted along the bond axis. The value of each pixel is determined as the sum of all function-values at the pixel
    position. Subsequently, the values are mapped to a color and drawn onto the canvas.

    Inspired from https://github.com/c-feldmann/rdkit_heatmaps/blob/master/rdkit_heatmaps/molmapping.py

    Parameters
    ----------
    mol: Chem.Mol
        RDKit molecule object which is displayed.
    grid_resolution: Sequence[int]
        Number of pixels of x- and y-axis.
    padding: Sequence[float]
        Increase of heatmap size, relative to size of molecule.
    atom_weights: Sequence[float] | npt.NDArray[np.float64] | None
        Array of weights for atoms.
    bond_weights: Sequence[float] | npt.NDArray[np.float64] | None
        Array of weights for bonds.
    atom_width: float
        Value for the width of displayed atom weights.
    bond_width: float
        Value for the width of displayed bond weights (perpendicular to bond-axis).
    bond_length: float
        Value for the length of displayed bond weights (along the bond-axis).

    Returns
    -------
    rdMolDraw2D.MolDraw2D
        Drawing of molecule and corresponding heatmap.
    """
    # assign default values and convert to numpy array
    if atom_weights is None:
        atom_weights = np.zeros(mol.GetNumAtoms())
    elif not isinstance(atom_weights, np.ndarray):
        atom_weights = np.array(atom_weights)

    if bond_weights is None:
        bond_weights = np.zeros(len(mol.GetBonds()))
    elif not isinstance(bond_weights, np.ndarray):
        bond_weights = np.array(bond_weights)

    # validate input
    if not len(atom_weights) == mol.GetNumAtoms():
        raise ValueError("len(atom_weights) is not equal to number of atoms in mol")

    if not len(bond_weights) == len(mol.GetBonds()):
        raise ValueError("len(bond_weights) is not equal to number of bonds in mol")

    # extract the 2D conformation of the molecule to be drawn
    conf = mol.GetConformer(0)

    # setup grid and add functions for atoms and bonds
    value_grid = _make_grid_from_mol(mol, grid_resolution, padding)
    value_grid = _add_gaussians_for_atoms(
        mol, conf, value_grid, atom_weights, atom_width
    )
    value_grid = _add_gaussians_for_bonds(
        mol, conf, value_grid, bond_weights, bond_width, bond_length
    )

    # evaluate all functions at pixel positions to obtain pixel values
    value_grid.evaluate()

    return value_grid


def _add_shap_present_absent_features_text(
    fig: Figure,
    explanation: SHAPFeatureAndAtomExplanation,
    sum_present_shap: float,
    sum_absent_shap: float,
) -> None:
    """Add text to the figure to display the SHAP prediction composition.

    The added text includes the prediction value, the expected value, the sum of the SHAP values for present features,
    and the sum of the SHAP values for absent features.

    Parameters
    ----------
    fig: Figure
        The figure.
    explanation: SHAPFeatureAndAtomExplanation
        The SHAP explanation.
    sum_present_shap: float
        The sum of the SHAP values for present features.
    sum_absent_shap: float
        The sum of the SHAP values for absent features.
    """
    if explanation.prediction is None:
        raise AssertionError("Prediction value is None.")
    if explanation.expected_value is None:
        raise AssertionError("Expected value is None.")

    color1 = "black"
    color2 = "green"
    color3 = "darkorchid"

    fontsize_numbers = 11
    delta = 0.04
    offset = 0.375
    fig.text(
        offset + delta,
        0.18,
        f"{explanation.prediction[-1]:.2f} =",
        fontsize=fontsize_numbers,
        ha="center",
    )
    fig.text(
        offset + 2 * delta,
        0.18,
        f" {'' if explanation.expected_value[-1] >= 0 else '-'}",
        ha="center",
        fontsize=fontsize_numbers,
        color=color1,
    )
    fig.text(
        offset + 3 * delta,
        0.18,
        f" {abs(explanation.expected_value[-1]):.2f}",
        ha="center",
        fontsize=fontsize_numbers,
        color=color1,
    )
    fig.text(
        offset + 4 * delta,
        0.18,
        f" {'+' if sum_present_shap >= 0 else '-'}",
        ha="center",
        fontsize=fontsize_numbers,
        color=color2,
    )
    fig.text(
        offset + 5 * delta,
        0.18,
        f" {abs(sum_present_shap):.2f}",
        ha="center",
        fontsize=fontsize_numbers,
        color=color2,
    )
    fig.text(
        offset + 6 * delta,
        0.18,
        f" {'+' if sum_absent_shap >= 0 else '-'}",
        fontsize=fontsize_numbers,
        ha="center",
        color=color3,
    )
    fig.text(
        offset + 7 * delta,
        0.18,
        f" {abs(sum_absent_shap):.2f}",
        ha="center",
        fontsize=fontsize_numbers,
        color=color3,
    )

    delta = 0.05
    offset = offset + 0.0165
    fig.text(offset, 0.13, "prediction =", ha="center", fontsize=10)
    fig.text(
        offset + 2 * delta,
        0.12,
        "expected\nvalue",
        ha="center",
        fontsize=10,
        color=color1,
    )
    fig.text(offset + 3 * delta, 0.13, " + ", ha="center", fontsize=10, color=color2)
    fig.text(
        offset + 4 * delta,
        0.12,
        "features\npresent",
        ha="center",
        fontsize=10,
        color=color2,
    )
    fig.text(offset + 5 * delta, 0.13, " + ", ha="center", fontsize=10, color=color3)
    fig.text(
        offset + 6 * delta,
        0.12,
        "features\nabsent",
        ha="center",
        fontsize=10,
        color=color3,
    )


def _structure_heatmap(
    mol: RDKitMol,
    atom_weights: npt.NDArray[np.float64],
    color: str | Colormap | tuple[RGBAtuple, RGBAtuple, RGBAtuple] | None = None,
    width: int = 600,
    height: int = 600,
    color_limits: tuple[float, float] | None = None,
) -> tuple[Draw.MolDraw2D, ValueGrid, ValueGrid, colors.Normalize, Colormap]:
    """Create a heatmap of the molecular structure, highlighting atoms with weighted Gaussian's.

    Parameters
    ----------
    mol: RDKitMol
        The molecule.
    atom_weights: npt.NDArray[np.float64]
        The atom weights.
    color: str | Colormap | tuple[RGBAtuple, RGBAtuple, RGBAtuple] | None
        The color map.
    width: int
        The width of the image in number of pixels.
    height: int
        The height of the image in number of pixels.
    color_limits: tuple[float, float] | None
        The color limits.

    Returns
    -------
    Draw.MolDraw2D, ValueGrid, ColorGrid, colors.Normalize, Colormap
        The configured drawer, the value grid, the color grid, the normalizer, and the
        color map.
    """
    drawer = Draw.MolDraw2DCairo(width, height)
    # Coloring atoms of element 0 to 100 black
    drawer.drawOptions().updateAtomPalette({i: (0, 0, 0, 1) for i in range(100)})
    draw_opt = drawer.drawOptions()
    draw_opt.padding = 0.2

    color_map = get_color_map_from_input(color)

    # create the sums of gaussians value grid
    mol_copy = Chem.Mol(mol)
    mol_copy = Draw.PrepareMolForDrawing(mol_copy)
    value_grid = make_sum_of_gaussians_grid(
        mol_copy,
        atom_weights=atom_weights,
        bond_weights=None,
        atom_width=0.5,  # 0.4
        bond_width=0.25,
        bond_length=0.5,
        grid_resolution=[drawer.Width(), drawer.Height()],
        padding=[draw_opt.padding * 2, draw_opt.padding * 2],
    )

    # create color-grid from the value grid.
    if color_limits is None:
        normalizer = get_color_normalizer_from_data(value_grid.values)
    else:
        normalizer = colors.Normalize(vmin=color_limits[0], vmax=color_limits[1])
    color_grid = value_grid.map2color(color_map, normalizer=normalizer)

    # draw the molecule and erase it to initialize the grid
    drawer.DrawMolecule(mol)
    drawer.ClearDrawing()
    # add the Colormap to the canvas
    color_canvas(drawer, color_grid)
    # add the molecule to the canvas
    drawer.DrawMolecule(mol)

    drawer.FinishDrawing()
    return drawer, value_grid, color_grid, normalizer, color_map


def structure_heatmap(
    mol: RDKitMol,
    atom_weights: npt.NDArray[np.float64],
    color: str | Colormap | tuple[RGBAtuple, RGBAtuple, RGBAtuple] | None = None,
    width: int = 600,
    height: int = 600,
    color_limits: tuple[float, float] | None = None,
) -> Image.Image:
    """Create a Gaussian plot on the molecular structure, highlight atoms with weighted Gaussians.

    Parameters
    ----------
    mol: RDKitMol
        The molecule.
    atom_weights: npt.NDArray[np.float64]
        The atom weights.
    color: str | Colormap | tuple[RGBAtuple, RGBAtuple, RGBAtuple] | None
        The color map.
    width: int
        The width of the image in number of pixels.
    height: int
        The height of the image in number of pixels.
    color_limits: tuple[float, float] | None
        The color limits.

    Returns
    -------
    Image
        The image as PNG.
    """
    drawer, *_ = _structure_heatmap(
        mol, atom_weights, color, width, height, color_limits
    )
    figure_bytes = drawer.GetDrawingText()
    image = to_png(figure_bytes)
    return image


def structure_heatmap_shap(  # pylint: disable=too-many-branches, too-many-locals
    explanation: SHAPFeatureAndAtomExplanation,
    color: str | Colormap | tuple[RGBAtuple, RGBAtuple, RGBAtuple] | None = None,
    width: int = 600,
    height: int = 600,
    color_limits: tuple[float, float] | None = None,
) -> Image.Image:
    """Create a heatmap of the molecular structure and display SHAP prediction composition.

    Parameters
    ----------
    explanation: SHAPExplanation
        The SHAP explanation.
    color: str | Colormap | tuple[RGBAtuple, RGBAtuple, RGBAtuple] | None
        The color map.
    width: int
        The width of the image in number of pixels.
    height: int
        The height of the image in number of pixels.
    color_limits: tuple[float, float] | None
        The color limits.

    Returns
    -------
    Image
        The image as PNG.
    """
    if explanation.feature_weights is None:
        raise ValueError("Explanation does not contain feature weights.")
    if explanation.feature_vector is None:
        raise ValueError("Explanation does not contain feature_vector.")
    if explanation.expected_value is None:
        raise ValueError("Explanation does not contain expected value.")
    if explanation.prediction is None:
        raise ValueError("Explanation does not contain prediction.")
    if explanation.molecule is None:
        raise ValueError("Explanation does not contain molecule.")
    if explanation.atom_weights is None:
        raise ValueError("Explanation does not contain atom weights.")

    if explanation.prediction.ndim > 2:
        raise ValueError(
            "Unsupported shape for prediction. Maximum 2 dimension is supported."
        )

    if explanation.feature_weights.ndim == 1:
        feature_weights = explanation.feature_weights
    elif explanation.feature_weights.ndim == 2:
        feature_weights = explanation.feature_weights[:, 1]
    else:
        raise ValueError("Unsupported shape for feature weights.")

    # calculate the sum of the SHAP values for present and absent features
    sum_present_shap, sum_absent_shap = calc_present_and_absent_shap_contributions(
        explanation.feature_vector, feature_weights
    )

    with plt.ioff():

        drawer, _, _, normalizer, color_map = _structure_heatmap(
            explanation.molecule,
            explanation.atom_weights,
            color=color,
            width=width,
            height=height,
            color_limits=color_limits,
        )
        figure_bytes = drawer.GetDrawingText()
        image_heatmap = to_png(figure_bytes)
        image_array = np.array(image_heatmap)

        fig, ax = plt.subplots(figsize=(8, 8))

        im = ax.imshow(
            image_array,
            cmap=color_map,
            norm=normalizer,
        )
        # remove ticks
        ax.set_xticks([])
        ax.set_yticks([])
        # remove border
        for spine in ax.spines.values():
            spine.set_visible(False)

        fig.colorbar(im, ax=ax, orientation="vertical", fraction=0.015, pad=0.0)

        _add_shap_present_absent_features_text(
            fig, explanation, sum_present_shap, sum_absent_shap
        )

        image = plt_to_pil(fig)
        # clear the figure and memory
        plt.close(fig)

        # remove dpi info because it crashes ipython's display function
        if "dpi" in image.info:
            del image.info["dpi"]
        # keep RDKit's image info
        image.info.update(image_heatmap.info)

    return image



"""Test utils for logging."""

from __future__ import annotations

from contextlib import contextmanager
from typing import Generator

import loguru
from loguru import logger


@contextmanager
def capture_logs(
    level: str = "INFO", log_format: str = "{level}:{name}:{message}"
) -> Generator[list[loguru.Message], None, None]:
    """Capture loguru-based logs.

    Custom context manager to test loguru-based logs. For details and usage examples,
    see https://loguru.readthedocs.io/en/latest/resources/migration.html#replacing-assertlogs-method-from-unittest-library

    Parameters
    ----------
    level : str, optional
        Log level, by default "INFO"
    log_format : str, optional
        Log format, by default "{level}:{name}:{message}"

    Yields
    -------
    list[loguru.Message]
        List of log messages

    Returns
    -------
    Generator[list[loguru.Message], None, None]
        List of log messages
    """
    output: list[loguru.Message] = []
    handler_id = logger.add(output.append, level=level, format=log_format)
    yield output
    logger.remove(handler_id)



"""Functions for counting the number of times a function is executed."""

from __future__ import annotations

from typing import Any

try:
    from typing import Self  # type: ignore[attr-defined]
except ImportError:
    from typing_extensions import Self

from sklearn.base import BaseEstimator
from sklearn.ensemble import RandomForestRegressor

from molpipeline import Pipeline
from molpipeline.abstract_pipeline_elements.core import ABCPipelineElement
from molpipeline.any2mol import SmilesToMol
from molpipeline.mol2any import MolToMorganFP


class CountingTransformerWrapper(BaseEstimator):
    """A transformer that counts the number of transformations."""

    def __init__(self, element: ABCPipelineElement):
        """Initialize the wrapper.

        Parameters
        ----------
        element : ABCPipelineElement
            The element to wrap.
        """
        self.element = element
        self.n_transformations = 0

    def fit(self, X: Any, y: Any) -> Self:  # pylint: disable=invalid-name
        """Fit the data.

        Parameters
        ----------
        X : Any
            The input data.
        y : Any
            The target data.

        Returns
        -------
        Any
            The fitted data.
        """
        self.element.fit(X, y)
        return self

    def transform(self, X: Any) -> Any:  # pylint: disable=invalid-name
        """Transform the data.

        Transform is called during prediction, which is not cached.
        Since the transformer is not cached, the counter is not increased.

        Parameters
        ----------
        X : Any
            The input data.

        Returns
        -------
        Any
            The transformed data.
        """
        return self.element.transform(X)

    def fit_transform(self, X: Any, y: Any) -> Any:  # pylint: disable=invalid-name
        """Fit and transform the data.

        Parameters
        ----------
        X : Any
            The input data.
        y : Any
            The target data.

        Returns
        -------
        Any
            The transformed data.
        """
        self.n_transformations += 1
        return self.element.fit_transform(X, y)

    def get_params(self, deep: bool = True) -> dict[str, Any]:
        """Get the parameters of the transformer.

        Parameters
        ----------
        deep : bool
            If True, the parameters of the transformer are also returned.

        Returns
        -------
        dict[str, Any]
            The parameters of the transformer.
        """
        params = {
            "element": self.element,
        }
        if deep:
            params.update(self.element.get_params(deep))
        return params

    def set_params(self, **params: Any) -> Self:
        """Set the parameters of the transformer.

        Parameters
        ----------
        **params
            The parameters to set.

        Returns
        -------
        Self
            The transformer with the set parameters
        """
        element = params.pop("element", None)
        if element is not None:
            self.element = element
        self.element.set_params(**params)
        return self


def get_exec_counted_rf_regressor(random_state: int) -> Pipeline:
    """Get a morgan + random forest pipeline, which counts the number of transformations.

    Parameters
    ----------
    random_state : int
        The random state to use.

    Returns
    -------
    Pipeline
        A pipeline with a morgan fingerprint, physchem descriptors, and a random forest
    """
    smi2mol = SmilesToMol()

    mol2concat = CountingTransformerWrapper(
        MolToMorganFP(radius=2, n_bits=2048),
    )
    rf = RandomForestRegressor(random_state=random_state, n_jobs=1)
    return Pipeline(
        [
            ("smi2mol", smi2mol),
            ("mol2concat", mol2concat),
            ("rf", rf),
        ],
        n_jobs=1,
    )



"""Test union find algorithm."""

import unittest

import numpy as np

from molpipeline.estimators.algorithm.union_find import UnionFindNode


class TestUnionFind(unittest.TestCase):
    """Test the UnionFindNode class."""

    def test_union_find(self) -> None:
        """Test the union find algorithm."""
        uf_nodes = [UnionFindNode() for _ in range(10)]
        nof_cc, cc_labels = UnionFindNode.get_connected_components(uf_nodes)

        self.assertTrue(np.equal(cc_labels, [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]).all())
        self.assertEqual(nof_cc, 10)

        uf_nodes = [UnionFindNode() for _ in range(10)]
        # one cc
        uf_nodes[0].union(uf_nodes[1])
        uf_nodes[0].union(uf_nodes[7])
        # second cc
        uf_nodes[2].union(uf_nodes[3])
        uf_nodes[2].union(uf_nodes[4])
        uf_nodes[3].union(uf_nodes[9])

        cc_num, cc_labels = UnionFindNode.get_connected_components(uf_nodes)
        self.assertTrue(
            np.equal(
                cc_labels,
                [
                    0,
                    0,
                    1,
                    1,
                    1,
                    2,
                    3,
                    0,
                    4,
                    1,
                ],
            ).all()
        )
        self.assertEqual(cc_num, 5)



"""Classes for encoding molecules as phys-chem vector."""

# pylint: disable=too-many-arguments

from __future__ import annotations

import warnings
from typing import Any, Iterable, Optional

import numpy as np
import numpy.typing as npt

try:
    from chemprop.data import MoleculeDatapoint, MoleculeDataset
    from chemprop.featurizers.base import GraphFeaturizer, VectorFeaturizer

    from molpipeline.estimators.chemprop.featurizer_wrapper.graph_wrapper import (
        SimpleMoleculeMolGraphFeaturizer,
    )
except ImportError:
    warnings.warn(
        "chemprop not installed. MolToChemprop will not work.",
        ImportWarning,
    )


from molpipeline.abstract_pipeline_elements.core import MolToAnyPipelineElement
from molpipeline.utils.molpipeline_types import RDKitMol


class MolToChemprop(MolToAnyPipelineElement):
    """PipelineElement for creating a graph representation used as input for Chemprop models.

    Each molecule is transformed to a MoleculeDatapoint object, which are then assembled into a MoleculeDataset.
    The MoleculeDataset can be used as input for Chemprop models.[1]

    References
    ----------
    [1] https://github.com/chemprop/chemprop/
    """

    graph_featurizer: GraphFeaturizer[RDKitMol] | None
    mol_featurizer: VectorFeaturizer[RDKitMol] | None

    def __init__(
        self,
        graph_featurizer: GraphFeaturizer[RDKitMol] | None = None,
        mol_featurizer: VectorFeaturizer[RDKitMol] | None = None,
        name: str = "Mol2Chemprop",
        n_jobs: int = 1,
        uuid: Optional[str] = None,
    ) -> None:
        """Initialize MolToChemprop.

        Parameters
        ----------
        graph_featurizer: GraphFeaturizer[RDKitMol] | None, optional (default=None)
            Defines how the graph is featurized. Defaults to None.
        mol_featurizer: MoleculeFeaturizer | None, optional (default=None)
            In contrast to graph_featurizer, features from the mol_featurizer are not used during the message passing.
            These features are concatenated to the neural fingerprints before the feedforward layers.
        name: str, optional (default="Mol2Chemprop")
            Name of the pipeline element. Defaults to "Mol2Chemprop".
        n_jobs: int
            Number of parallel jobs to use. Defaults to 1.
        uuid: str | None, optional (default=None)
            UUID of the pipeline element.
        """
        self.graph_featurizer = graph_featurizer or SimpleMoleculeMolGraphFeaturizer()
        self.mol_featurizer = mol_featurizer

        super().__init__(
            name=name,
            n_jobs=n_jobs,
            uuid=uuid,
        )

    def pretransform_single(self, value: RDKitMol) -> MoleculeDatapoint:
        """Transform a single molecule to a ChemProp MoleculeDataPoint.

        Parameters
        ----------
        value: RDKitMol
            RDKit molecule to transform.

        Returns
        -------
        MoleculeDatapoint
            Molecular representation used as input for ChemProp. None if transformation failed.
        """
        mol_features: npt.NDArray[np.float64] | None = None
        if self.mol_featurizer is not None:
            mol_features = np.array(self.mol_featurizer(value), dtype=np.float64)
        return MoleculeDatapoint(mol=value, x_d=mol_features)

    def assemble_output(
        self, value_list: Iterable[MoleculeDatapoint]
    ) -> MoleculeDataset:
        """Assemble the output from the parallelized pretransform_single.

        Parameters
        ----------
        value_list: Iterable
            List of transformed values.

        Returns
        -------
        Any
            Assembled output.
        """
        return MoleculeDataset(data=list(value_list), featurizer=self.graph_featurizer)

    def get_params(self, deep: bool = True) -> dict[str, Any]:
        """Get parameters for this pipeline element.

        Parameters
        ----------
        deep: bool, optional (default=True)
            If True, will return the parameters for this pipeline element and its subobjects.

        Returns
        -------
        dict
            Parameters of this pipeline element.
        """
        params = super().get_params(deep=deep)
        params["graph_featurizer"] = self.graph_featurizer
        params["mol_featurizer"] = self.mol_featurizer

        if deep:
            if hasattr(self.graph_featurizer, "get_params"):
                graph_featurizer_params = self.graph_featurizer.get_params(deep=deep)  # type: ignore
                for key, value in graph_featurizer_params.items():
                    params[f"graph_featurizer__{key}"] = value
            if hasattr(self.mol_featurizer, "get_params"):
                mol_featurizer_params = self.mol_featurizer.get_params(deep=deep)  # type: ignore
                for key, value in mol_featurizer_params.items():
                    params[f"mol_featurizer__{key}"] = value
        return params

    def set_params(self, **parameters: Any) -> MolToChemprop:
        """Set the parameters of this pipeline element.

        Parameters
        ----------
        **parameters: Any
            Parameters to set.

        Returns
        -------
        MolToChemprop
            This pipeline element with the parameters set.
        """
        param_copy = dict(parameters)
        graph_featurizer = param_copy.pop("graph_featurizer", None)
        mol_featurizer = param_copy.pop("mol_featurizer", None)
        if graph_featurizer is not None:
            self.graph_featurizer = graph_featurizer
        if mol_featurizer is not None:
            self.mol_featurizer = mol_featurizer
        graph_featurizer_params = {}
        mol_featurizer_params = {}
        for key in list(param_copy.keys()):
            if "__" not in key:
                continue
            component_name, _, param_name = key.partition("__")
            if component_name == "graph_featurizer":
                graph_featurizer_params[param_name] = param_copy.pop(key)
            elif component_name == "mol_featurizer":
                mol_featurizer_params[param_name] = param_copy.pop(key)
        if hasattr(self.graph_featurizer, "set_params"):
            self.graph_featurizer.set_params(**graph_featurizer_params)  # type: ignore
        if hasattr(self.mol_featurizer, "set_params"):
            self.mol_featurizer.set_params(**mol_featurizer_params)  # type: ignore

        super().set_params(**param_copy)
        return self



"""Functions for serializing and deserializing PyTorch models."""

from typing import TypeVar

try:
    import torch

    TORCH_AVAILABLE = True
except ImportError:
    TORCH_AVAILABLE = False
from typing import Any, Literal

_T = TypeVar("_T")

if TORCH_AVAILABLE:

    def tensor_to_json(
        obj: _T,
    ) -> tuple[dict[str, Any], Literal[True]] | tuple[_T, Literal[False]]:
        """Recursively convert a PyTorch model to a JSON-serializable object.

        Parameters
        ----------
        obj : object
            The object to convert.

        Returns
        -------
        object
            The JSON-serializable object.
        """
        if isinstance(obj, torch.Tensor):
            object_dict: dict[str, Any] = {
                "__name__": obj.__class__.__name__,
                "__module__": obj.__class__.__module__,
                "__init__": True,
            }
        else:
            return obj, False
        object_dict["data"] = obj.tolist()
        return object_dict, True

else:

    def tensor_to_json(
        obj: _T,
    ) -> tuple[dict[str, Any], Literal[True]] | tuple[_T, Literal[False]]:
        """Recursively convert a PyTorch model to a JSON-serializable object.

        Parameters
        ----------
        obj : object
            The object to convert.

        Returns
        -------
        object
            The JSON-serializable object.
        """
        return obj, False



"""Converter element for molecules to binary string representation."""

from rdkit import Chem

from molpipeline.abstract_pipeline_elements.core import MolToAnyPipelineElement


class MolToBinary(MolToAnyPipelineElement):
    """PipelineElement to transform a molecule to a binary."""

    def pretransform_single(self, value: Chem.Mol) -> str:
        """Transform a molecule to a binary string.

        Parameters
        ----------
        value: Chem.Mol
            Molecule to be transformed to binary string representation.

        Returns
        -------
        str
            Binary representation of molecule.
        """
        return value.ToBinary()



"""Parent class of wrapper for Chemprop to make it compatible with scikit-learn."""

import abc
from typing import Any, Sequence

# pylint: disable=duplicate-code
try:
    from typing import Self  # type: ignore[attr-defined]
except ImportError:
    from typing_extensions import Self

import numpy as np
import numpy.typing as npt

try:
    from chemprop.data import MoleculeDataset, build_dataloader
    from chemprop.models.model import MPNN
    from lightning import pytorch as pl
except ImportError:
    pass
from sklearn.base import BaseEstimator

from molpipeline.estimators.chemprop.lightning_wrapper import get_params_trainer

# pylint: enable=duplicate-code


class ABCChemprop(BaseEstimator, abc.ABC):
    """Wrap Chemprop in a sklearn compatible Estimator.

    Notes
    -----
    As the ChempropNeuralFP is a transformer and the ChempropModel is a predictor, this class
    provides the common functionality for both classes.
    Although this class does not implement abstract methods, it is marked as abstract to prevent
    instantiation. (without predict or transform methods this class is useless.)

    Attributes
    ----------
    model : MPNN
        The chemprop model to wrap.
    batch_size : int
        The batch size to use for training and prediction.
    n_jobs : int
        The number of jobs to use for processing the molecular graphs in the dataloader.
    lightning_trainer : pl.Trainer
        The lightning trainer to use for training the model.
    trainer_params : dict[str, Any]
        The parameters of the lightning trainer. This is used as the trainer is not compatible with the
        `get_params` method.
    model_ckpoint_params : dict[str, Any]
        The parameters of the model checkpoint callback. This is used as the callback is not compatible with the
        `get_params` method.
    """

    model: MPNN
    batch_size: int
    n_jobs: int
    lightning_trainer: pl.Trainer
    trainer_params: dict[str, Any]
    model_ckpoint_params: dict[str, Any]

    def __init__(
        self,
        model: MPNN,
        lightning_trainer: pl.Trainer | None = None,
        batch_size: int = 64,
        n_jobs: int = 1,
        **kwargs: Any,
    ) -> None:
        """Initialize the chemprop abstract model.

        Parameters
        ----------
        model : MPNN
            The chemprop model to wrap.
        lightning_trainer : pl.Trainer, optional
            The lightning trainer to use, by default None
        batch_size : int, optional (default=64)
            The batch size to use.
        n_jobs : int, optional (default=1)
            The number of jobs to use.
        kwargs : Any
            Parameters set using `set_params`.
            Can be used to modify components of the model.
        """
        self.model = model
        self.batch_size = batch_size
        self.n_jobs = n_jobs
        self.model_ckpoint_params = {}
        if not lightning_trainer:
            lightning_trainer = pl.Trainer(
                logger=False,
                enable_checkpointing=False,
                max_epochs=500,
                enable_model_summary=False,
                callbacks=[],
            )
        self.lightning_trainer = lightning_trainer
        self.trainer_params = get_params_trainer(self.lightning_trainer)
        self.set_params(**kwargs)

    def _update_trainer(
        self,
    ) -> None:
        """Update the trainer for the model."""
        trainer_params = dict(self.trainer_params)
        if self.model_ckpoint_params:
            trainer_params["callbacks"] = [
                pl.callbacks.ModelCheckpoint(**self.model_ckpoint_params)
            ]
        self.lightning_trainer = pl.Trainer(**trainer_params)

    def fit(
        self,
        X: MoleculeDataset,  # pylint: disable=invalid-name
        y: Sequence[int | float] | npt.NDArray[np.int_ | np.float64],
    ) -> Self:
        """Fit the model to the data.

        Parameters
        ----------
        X : MoleculeDataset
            The input data.
        y : Sequence[int | float] | npt.NDArray[np.int_ | np.float64]
            The target data.

        Returns
        -------
        Self
            The fitted model.
        """
        if not isinstance(y, np.ndarray):
            y = np.array(y)
        if y.ndim == 1:
            y = y.reshape(-1, 1)
        X.Y = y
        training_data = build_dataloader(
            X, batch_size=self.batch_size, num_workers=self.n_jobs
        )
        self.lightning_trainer.fit(self.model, training_data)
        return self

    def set_params(self, **params: Any) -> Self:
        """Set the parameters of the model.

        Note
        ----
        Parameters for the trainer and the checkpoint callback are filtered out and added as attributes of the model.
        This is done due to incompatibility with the `get_params` method.

        Parameters
        ----------
        **params: Any
            The parameters to set.

        Returns
        -------
        Self
            The model with the new parameters.
        """
        params, trainer_params = self._filter_params(params, "lightning_trainer")
        params, model_ckpoint_params = self._filter_params(params, "callback_modelckpt")
        self.trainer_params.update(trainer_params)
        self.model_ckpoint_params.update(model_ckpoint_params)
        super().set_params(**params)
        self._update_trainer()
        return self

    def get_params(self, deep: bool = False) -> dict[str, Any]:
        """Get the parameters of the model.

        Parameters
        ----------
        deep : bool, optional (default=False)
            Whether to get the parameters of the model.

        Notes
        -----
        The parameters of the trainer and the model checkpoint are added irrespective of the `deep` parameter.
        This is done due to their incompatibility with the `get_params` and `set_params` methods.

        Returns
        -------
        dict[str, Any]
            The parameters of the model.
        """
        params = super().get_params(deep)
        # Since the trainer and the model checkpoint are not compatible with the `get_params` and `set_params` methods
        # they are not passed as objects but as parameters. Hence, the `deep` parameter is ignored and the parameters
        # are always returned.
        for name, value in self.trainer_params.items():
            params[f"lightning_trainer__{name}"] = value
        for name, value in self.model_ckpoint_params.items():
            params[f"callback_modelckpt__{name}"] = value
        # set to none as the trainer is created from the parameters
        # Otherwise, the sklearn clone will fail as the trainer is updated by replacing the object
        params["lightning_trainer"] = None
        return params

    @staticmethod
    def _filter_params(
        params: dict[str, Any],
        prefix: str,
    ) -> tuple[dict[str, Any], dict[str, Any]]:
        """Filter the parameters for the specified prefix.

        Parameters
        ----------
        params : dict[str, Any]
            The parameters to filter.
        prefix : str
            The prefix to filter the parameters for.

        Returns
        -------
        dict[str, Any]
            The filtered parameters for the model.
        dict[str, Any]
            The filtered parameters for the trainer.
        """
        trainer_params = {}
        other_params = {}
        for key, value in params.items():
            if key.startswith(f"{prefix}__"):
                trainer_params[key.split("__")[1]] = value
            else:
                other_params[key] = value
        return other_params, trainer_params



"""Scorer that ignores a given value in the prediction array."""

from __future__ import annotations

from typing import Any

import numpy as np
import numpy.typing as npt
import pandas as pd
from loguru import logger
from sklearn import metrics
from sklearn.metrics._scorer import _BaseScorer  # pylint: disable=protected-access


def ignored_value_scorer(
    scorer: str | _BaseScorer, ignore_value: Any = None
) -> _BaseScorer:
    """Create a scorer that ignores a given value in the prediction array.

    This is relevant for pipline models which replace errors with a given value.
    The wrapped scorer will ignore that value and return the corresponding score.

    Parameters
    ----------
    scorer : str or _BaseScorer
        The scorer to wrap.
    ignore_value : Any, optional
        The value to ignore in the prediction array.
        Default: None

    Returns
    -------
    _BaseScorer
        The scorer that ignores the given value.
    """
    if isinstance(scorer, str):
        scorer = metrics.get_scorer(scorer)

    score_func = scorer._score_func  # pylint: disable=protected-access
    response_method = scorer._response_method  # pylint: disable=protected-access
    scorer_kwargs = scorer._kwargs  # pylint: disable=protected-access
    if scorer._sign < 0:  # pylint: disable=protected-access
        scorer_kwargs["greater_is_better"] = False

    def newscore(
        y_true: npt.NDArray[np.float64 | np.int_],
        y_pred: npt.NDArray[np.float64 | np.int_],
        **kwargs: Any,
    ) -> float:
        """Compute the score for the given prediction arrays.

        Parameters
        ----------
        y_true : npt.NDArray[np.float64 | np.int_]
            The true values.
        y_pred : npt.NDArray[np.float64 | np.int_]
            The predicted values.
        **kwargs
            Additional keyword arguments.

        Returns
        -------
        float
            The score for the given prediction arrays.
        """
        retained_y_true: npt.NDArray[np.bool_]
        retained_y_pred: npt.NDArray[np.bool_]
        if pd.notna(ignore_value):
            retained_y_true = ~np.equal(y_true, ignore_value)
            retained_y_pred = ~np.equal(y_pred, ignore_value)
        else:
            retained_y_true = pd.notna(y_true)
            retained_y_pred = pd.notna(y_pred)

        all_retained = retained_y_pred & retained_y_true

        if not np.all(all_retained):
            logger.warning(
                f"Warning, prediction array contains NaN values, removing {sum(~all_retained)} elements"
            )
        y_true_ = np.copy(np.array(y_true)[all_retained])
        y_pred_ = np.array(np.array(y_pred)[all_retained].tolist())
        _kwargs = dict(kwargs)
        if "sample_weight" in _kwargs and _kwargs["sample_weight"] is not None:
            _kwargs["sample_weight"] = _kwargs["sample_weight"][all_retained]
        return score_func(y_true_, y_pred_, **_kwargs)

    return metrics.make_scorer(
        newscore, response_method=response_method, **scorer_kwargs
    )



"""Pipline elements for mol to mol transformations."""

# pylint: disable=too-many-arguments

from __future__ import annotations

from typing import Any, Literal, Optional

try:
    from typing import Self  # type: ignore[attr-defined]
except ImportError:
    from typing_extensions import Self

import copy
import warnings

from rdkit.Chem import AllChem

from molpipeline.abstract_pipeline_elements.core import (
    InvalidInstance,
    MolToMolPipelineElement,
)
from molpipeline.utils.molpipeline_types import OptionalMol, RDKitMol


class MolToMolReaction(MolToMolPipelineElement):
    """PipelineElement which transforms the input according to the specified reaction."""

    additive_list: list[RDKitMol]
    handle_multi: Literal["pass", "warn", "raise"]
    _reaction: AllChem.ChemicalReaction

    def __init__(
        self,
        reaction: AllChem.ChemicalReaction,
        additive_list: list[RDKitMol],
        handle_multi: Literal["pass", "warn", "raise"] = "warn",
        name: str = "MolToMolReaction",
        n_jobs: int = 1,
        uuid: Optional[str] = None,
    ) -> None:
        """Initialize MolToMolReaction.

        Parameters
        ----------
        reaction: AllChem.ChemicalReaction
            Reaction which is applied to input.
        additive_list: list[Chem.Mol]
            Molecules which are added as educts to the reaction, but are not part of input.
        handle_multi: Literal["pass", "warn", "raise"]
            How to handle reaction where multiple products are possible.
        name: str, optional (default="MolToMolReaction")
            Name of PipelineElement.
        n_jobs: int, optional (default=1)
            Number of cores used.
        uuid: str | None, optional (default=None)
            UUID of the pipeline element. If None, a random UUID is generated.
        """
        super().__init__(
            name=name,
            n_jobs=n_jobs,
            uuid=uuid,
        )
        self.reaction = reaction
        self.additive_list = additive_list
        self.handle_multi = handle_multi

    def get_params(self, deep: bool = True) -> dict[str, Any]:
        """Return all parameters defining the object.

        Parameters
        ----------
        deep: bool
            If True get a deep copy of the parameters.

        Returns
        -------
        dict[str, Any]
            Dictionary containing all parameters defining the object.
        """
        parameters = super().get_params(deep)
        if deep:
            parameters["reaction"] = AllChem.ChemicalReaction(self.reaction)
            parameters["additive_list"] = [
                RDKitMol(additive) for additive in self.additive_list
            ]
            parameters["handle_multi"] = copy.copy(self.handle_multi)
        else:
            parameters["reaction"] = self.reaction
            parameters["additive_list"] = self.additive_list
            parameters["handle_multi"] = self.handle_multi
        return parameters

    def set_params(self, **parameters: Any) -> Self:
        """Set the parameters.

        Parameters
        ----------
        parameters: Any
            Dictionary containing parameters to be set.

        Returns
        -------
        Self
            MolToMolReaction with updated parameters.
        """
        super().set_params(**parameters)
        if "reaction" in parameters:
            self.reaction = parameters["reaction"]
        if "additive_list" in parameters:
            self.additive_list = parameters["additive_list"]
        if "handle_multi" in parameters:
            self.handle_multi = parameters["handle_multi"]
        return self

    @property
    def reaction(self) -> AllChem.ChemicalReaction:
        """Get the reaction which is applied to the input molecule."""
        return self._reaction

    @reaction.setter
    def reaction(self, reaction: AllChem.ChemicalReaction) -> None:
        """Set the reaction which is applied to the input molecule.

        Parameters
        ----------
        reaction: AllChem.ChemicalReaction
            Reaction which is applied to molecules.

        Returns
        -------
        None
        """
        if not isinstance(reaction, AllChem.ChemicalReaction):
            raise TypeError("Not a Chemical reaction!")
        self._reaction = reaction

    def pretransform_single(self, value: RDKitMol) -> OptionalMol:
        """Apply reaction to molecule.

        Parameters
        ----------
        value: RDKitMol
            Molecule to apply reaction to.

        Returns
        -------
        OptionalMol
            Product of reaction if possible, else InvalidInstance.
        """
        mol = value  # Only value to keep signature consistent.
        reactant_list: list[RDKitMol] = list(self.additive_list)
        reactant_list.append(mol)
        product_list = self.reaction.RunReactants(reactant_list)

        if len(product_list) > 1:
            if self.handle_multi == "warn":
                warnings.warn(
                    "Not able to handle multiple reactions. An arbitrary reaction is selected."
                )
            elif self.handle_multi == "raise":
                if mol.HasProp("identifier"):
                    mol_id = mol.GetProp("identifier")
                else:
                    mol_id = "None"
                raise ValueError(
                    f"Not able to handle multiple reactions: Mol ID: {mol_id}"
                )

        if len(product_list) == 0:
            return InvalidInstance(
                self.uuid, "Reaction did not yield any product.", self.name
            )
        product = product_list[0][0]
        AllChem.SanitizeMol(product)  # type: ignore[attr-defined]
        return product



"""Utility functions for multiprocessing."""

from __future__ import annotations

import multiprocessing
import warnings


def check_available_cores(n_requested_cores: int) -> int:
    """Compare number of requested cores with available cores and return a (corrected) number.

    Parameters
    ----------
    n_requested_cores: int
        Number of requested cores.

    Returns
    -------
    int
        Number of used cores.
    """
    if not isinstance(n_requested_cores, int):
        raise TypeError(f"Not an integer: {n_requested_cores}")
    try:
        n_available_cores = multiprocessing.cpu_count()
    except ModuleNotFoundError:
        warnings.warn(
            "Cannot import multiprocessing library. Falling back to single core!"
        )
        return 1

    if n_requested_cores > n_available_cores:
        warnings.warn(
            "Requested more cores than available. Using maximum number of cores!"
        )
        return n_available_cores
    if n_requested_cores < 0:
        return n_available_cores

    return n_requested_cores



"""Test functionality of the pipeline class."""

from __future__ import annotations

import tempfile
import unittest
from itertools import combinations
from pathlib import Path
from typing import Any

import numpy as np
import pandas as pd
from joblib import Memory
from sklearn.base import BaseEstimator
from sklearn.calibration import CalibratedClassifierCV
from sklearn.ensemble import RandomForestClassifier, RandomForestRegressor
from sklearn.model_selection import GridSearchCV
from sklearn.tree import DecisionTreeClassifier

from molpipeline import ErrorFilter, FilterReinserter, Pipeline, PostPredictionWrapper
from molpipeline.any2mol import AutoToMol, SmilesToMol
from molpipeline.mol2any import MolToMorganFP, MolToRDKitPhysChem, MolToSmiles
from molpipeline.mol2mol import (
    ChargeParentExtractor,
    EmptyMoleculeFilter,
    MetalDisconnector,
    SaltRemover,
)
from molpipeline.utils.json_operations import recursive_from_json, recursive_to_json
from molpipeline.utils.matrices import are_equal
from tests import TEST_DATA_DIR
from tests.utils.execution_count import get_exec_counted_rf_regressor
from tests.utils.fingerprints import make_sparse_fp

TEST_SMILES = ["CC", "CCO", "COC", "CCCCC", "CCC(-O)O", "CCCN"]
FAULTY_TEST_SMILES = ["CCCXAS", "", "O=C(O)C(F)(F)F"]
CONTAINS_OX = [0, 1, 1, 0, 1, 0]
FP_RADIUS = 2
FP_SIZE = 2048
EXPECTED_OUTPUT = make_sparse_fp(TEST_SMILES, FP_RADIUS, FP_SIZE)

_RANDOM_STATE = 67056


class PipelineTest(unittest.TestCase):
    """Unit test for the functionality of the pipeline class."""

    def test_fit_transform_single_core(self) -> None:
        """Test if the generation of the fingerprint matrix works as expected.

        Returns
        -------
        None
        """
        # Create pipeline
        smi2mol = SmilesToMol()
        mol2morgan = MolToMorganFP(radius=FP_RADIUS, n_bits=FP_SIZE)
        pipeline = Pipeline(
            [
                ("smi2mol", smi2mol),
                ("morgan", mol2morgan),
            ]
        )

        # Run pipeline
        matrix = pipeline.fit_transform(TEST_SMILES)

        # Compare with expected output
        self.assertTrue(are_equal(EXPECTED_OUTPUT, matrix))

    def test_sklearn_pipeline(self) -> None:
        """Test if the pipeline can be used in a sklearn pipeline.

        Returns
        -------
        None
        """
        smi2mol = SmilesToMol()
        mol2morgan = MolToMorganFP(radius=FP_RADIUS, n_bits=FP_SIZE)
        d_tree = DecisionTreeClassifier()
        s_pipeline = Pipeline(
            [
                ("smi2mol", smi2mol),
                ("morgan", mol2morgan),
                ("decision_tree", d_tree),
            ]
        )
        s_pipeline.fit(TEST_SMILES, CONTAINS_OX)
        predicted_value_array = s_pipeline.predict(TEST_SMILES)
        for pred_val, true_val in zip(predicted_value_array, CONTAINS_OX):
            self.assertEqual(pred_val, true_val)

    def test_sklearn_pipeline_parallel(self) -> None:
        """Test if the pipeline can be used in a sklearn pipeline.

        Returns
        -------
        None
        """
        smi2mol = SmilesToMol()
        mol2morgan = MolToMorganFP(radius=FP_RADIUS, n_bits=FP_SIZE)
        d_tree = DecisionTreeClassifier()
        s_pipeline = Pipeline(
            [
                ("smi2mol", smi2mol),
                ("morgan", mol2morgan),
                ("decision_tree", d_tree),
            ],
            n_jobs=2,
        )
        s_pipeline.fit(TEST_SMILES, CONTAINS_OX)
        out = s_pipeline.predict(TEST_SMILES)
        self.assertEqual(len(out), len(CONTAINS_OX))
        for pred_val, true_val in zip(out, CONTAINS_OX):
            self.assertEqual(pred_val, true_val)

    def test_salt_removal(self) -> None:
        """Test if salts are correctly removed from molecules.

        Returns
        -------
        None
        """
        smiles_with_salt_list = ["CCO-[Na]", "CCC(=O)[O-].[Li+]", "CCC(=O)-O-[K]"]
        smiles_without_salt_list = ["CCO", "CCC(=O)O", "CCC(=O)O"]

        smi2mol = SmilesToMol()
        disconnect_metal = MetalDisconnector()
        salt_remover = SaltRemover()
        empty_mol_filter = EmptyMoleculeFilter()
        remove_charge = ChargeParentExtractor()
        mol2smi = MolToSmiles()

        salt_remover_pipeline = Pipeline(
            [
                ("smi2mol", smi2mol),
                ("disconnect_metal", disconnect_metal),
                ("salt_remover", salt_remover),
                ("empty_mol_filter", empty_mol_filter),
                ("remove_charge", remove_charge),
                ("mol2smi", mol2smi),
            ]
        )
        generated_smiles = salt_remover_pipeline.transform(smiles_with_salt_list)
        for generated_smiles, smiles_without_salt in zip(
            generated_smiles, smiles_without_salt_list
        ):
            self.assertEqual(generated_smiles, smiles_without_salt)

    def test_json_generation(self) -> None:
        """Test that the json representation of a pipeline can be loaded back into a pipeline.

        Returns
        -------
        None
        """

        # Create pipeline
        smi2mol = SmilesToMol()
        metal_disconnector = MetalDisconnector()
        salt_remover = SaltRemover()
        physchem = MolToRDKitPhysChem()
        pipeline_element_list = [
            smi2mol,
            metal_disconnector,
            salt_remover,
            physchem,
        ]
        m_pipeline = Pipeline(
            [
                ("smi2mol", smi2mol),
                ("metal_disconnector", metal_disconnector),
                ("salt_remover", salt_remover),
                ("physchem", physchem),
            ]
        )

        # Convert pipeline to json
        json_str = recursive_to_json(m_pipeline)
        # Recreate pipeline from json
        loaded_pipeline: Pipeline = recursive_from_json(json_str)
        self.assertTrue(isinstance(loaded_pipeline, Pipeline))
        # Compare pipeline elements
        for loaded_element, original_element in zip(
            loaded_pipeline.steps, pipeline_element_list
        ):
            if loaded_element[1] == "passthrough":
                self.assertEqual(loaded_element[1], original_element)
                continue
            loaded_params = loaded_element[1].get_params()
            original_params = original_element.get_params()
            for key, value in loaded_params.items():
                if isinstance(value, BaseEstimator):
                    self.assertEqual(type(value), type(original_params[key]))
                else:
                    self.assertEqual(loaded_params[key], original_params[key])

    def test_fit_transform_record_remove_nones(self) -> None:
        """Test if the generation of the fingerprint matrix works as expected.
        Returns
        -------
        None
        """
        smi2mol = SmilesToMol()
        salt_remover = SaltRemover()
        mol2morgan = MolToMorganFP(radius=FP_RADIUS, n_bits=FP_SIZE)
        empty_mol_filter = EmptyMoleculeFilter()
        remove_none = ErrorFilter.from_element_list(
            [smi2mol, salt_remover, mol2morgan, empty_mol_filter]
        )
        # Create pipeline
        pipeline = Pipeline(
            [
                ("smi2mol", smi2mol),
                ("salt_remover", salt_remover),
                ("empty_mol_filter", empty_mol_filter),
                ("morgan", mol2morgan),
                ("remove_none", remove_none),
            ],
        )

        # Run pipeline
        matrix = pipeline.fit_transform(TEST_SMILES + FAULTY_TEST_SMILES)
        # Compare with expected output (Which is the same as the output without the faulty smiles)
        self.assertTrue(are_equal(EXPECTED_OUTPUT, matrix))

    def test_caching(self) -> None:
        """Test if the caching gives the same results and is faster on the second run."""

        molecule_net_logd_df = pd.read_csv(
            TEST_DATA_DIR / "molecule_net_logd.tsv.gz", sep="\t", nrows=20
        )
        prediction_list = []
        for cache_activated in [False, True]:
            pipeline = get_exec_counted_rf_regressor(_RANDOM_STATE)
            with tempfile.TemporaryDirectory() as temp_dir:

                if cache_activated:
                    cache_dir = Path(temp_dir) / ".cache"
                    mem = Memory(location=cache_dir, verbose=0)
                else:
                    mem = Memory(location=None, verbose=0)
                pipeline.memory = mem
                # Run fitting 1
                pipeline.fit(
                    molecule_net_logd_df["smiles"].tolist(),
                    molecule_net_logd_df["exp"].tolist(),
                )
                # Get predictions
                prediction = pipeline.predict(molecule_net_logd_df["smiles"].tolist())
                prediction_list.append(prediction)

                # Reset the last step with an untrained model
                pipeline.steps[-1] = (
                    "rf",
                    RandomForestRegressor(random_state=_RANDOM_STATE, n_jobs=1),
                )

                # Run fitting 2
                pipeline.fit(
                    molecule_net_logd_df["smiles"].tolist(),
                    molecule_net_logd_df["exp"].tolist(),
                )
                # Get predictions
                prediction = pipeline.predict(molecule_net_logd_df["smiles"].tolist())
                prediction_list.append(prediction)

                n_transformations = pipeline.named_steps["mol2concat"].n_transformations
                if cache_activated:
                    # Fit is called twice, but the transform is only called once, since the second run is cached
                    self.assertEqual(n_transformations, 1)
                else:
                    self.assertEqual(n_transformations, 2)

                mem.clear(warn=False)
            for pred1, pred2 in combinations(prediction_list, 2):
                self.assertTrue(np.allclose(pred1, pred2))


class PipelineCompatibilityTest(unittest.TestCase):
    """Test if the pipeline is compatible with other sklearn functionalities."""

    def test_gridsearchcv(self) -> None:
        """Test if the MolPipeline can be used in sklearn's GridSearchCV."""

        descriptor_elements_to_test: list[dict[str, Any]] = [
            {
                "name": "morgan",
                "element": MolToMorganFP(),
                "param_grid": {"morgan__n_bits": [64, 128], "morgan__radius": [1, 2]},
            },
            {
                "name": "physchem",
                "element": MolToRDKitPhysChem(),
                "param_grid": {
                    "physchem__descriptor_list": [
                        ["HeavyAtomMolWt"],
                        ["HeavyAtomMolWt", "HeavyAtomCount"],
                    ]
                },
            },
        ]

        for test_data_dict in descriptor_elements_to_test:

            name = test_data_dict["name"]
            element = test_data_dict["element"]
            param_grid = test_data_dict["param_grid"]

            # set up a pipeline that trains a random forest classifier on morgan fingerprints
            pipeline = Pipeline(
                [
                    ("auto2mol", AutoToMol()),
                    (name, element),
                    ("estimator", RandomForestClassifier()),
                ],
                n_jobs=1,
            )

            # define the hyperparameter space to try out
            grid_space = {
                "estimator__n_estimators": [1, 5],
                "estimator__random_state": [_RANDOM_STATE],
            }
            grid_space.update(param_grid)

            grid_search_cv = GridSearchCV(
                estimator=pipeline,
                param_grid=grid_space,
                cv=2,
                scoring="roc_auc",
                n_jobs=1,
            )

            grid_search_cv.fit(
                X=TEST_SMILES,
                y=CONTAINS_OX,
            )

            for k, value in param_grid.items():
                self.assertIn(grid_search_cv.best_params_[k], value)

    def test_gridsearch_cache(self) -> None:
        """Run a short GridSearchCV and check if the caching and not caching gives the same results."""
        h_params = {
            "rf__n_estimators": [1, 2],
        }
        # First without caching
        data_df = pd.read_csv(
            TEST_DATA_DIR / "molecule_net_logd.tsv.gz", sep="\t", nrows=20
        )
        best_param_dict = {}
        prediction_dict = {}
        for cache_activated in [True, False]:
            pipeline = get_exec_counted_rf_regressor(_RANDOM_STATE)
            with tempfile.TemporaryDirectory() as temp_dir:
                cache_dir = Path(temp_dir) / ".cache"
                if cache_activated:
                    mem = Memory(location=cache_dir, verbose=0)
                else:
                    mem = Memory(location=None, verbose=0)
                pipeline.memory = mem
                grid_search_cv = GridSearchCV(
                    estimator=pipeline,
                    param_grid=h_params,
                    cv=2,
                    scoring="neg_mean_squared_error",
                    n_jobs=1,
                    error_score="raise",
                    refit=True,
                    pre_dispatch=1,
                )
                grid_search_cv.fit(data_df["smiles"].tolist(), data_df["exp"].tolist())
                best_param_dict[cache_activated] = grid_search_cv.best_params_
                prediction_dict[cache_activated] = grid_search_cv.predict(
                    data_df["smiles"].tolist()
                )
                mem.clear(warn=False)
        self.assertEqual(best_param_dict[True], best_param_dict[False])
        self.assertTrue(np.allclose(prediction_dict[True], prediction_dict[False]))

    def test_calibrated_classifier(self) -> None:
        """Test if the pipeline can be used with a CalibratedClassifierCV."""
        smi2mol = SmilesToMol()
        mol2morgan = MolToMorganFP(radius=FP_RADIUS, n_bits=FP_SIZE)
        d_tree = DecisionTreeClassifier()
        error_filter = ErrorFilter(filter_everything=True)
        s_pipeline = Pipeline(
            [
                ("smi2mol", smi2mol),
                ("morgan", mol2morgan),
                ("error_filter", error_filter),
                ("decision_tree", d_tree),
                (
                    "error_replacer",
                    PostPredictionWrapper(
                        FilterReinserter.from_error_filter(error_filter, np.nan)
                    ),
                ),
            ]
        )
        calibrated_pipeline = CalibratedClassifierCV(
            s_pipeline, cv=2, ensemble=True, method="isotonic"
        )
        calibrated_pipeline.fit(TEST_SMILES, CONTAINS_OX)
        predicted_value_array = calibrated_pipeline.predict(TEST_SMILES)
        predicted_proba_array = calibrated_pipeline.predict_proba(TEST_SMILES)
        self.assertIsInstance(predicted_value_array, np.ndarray)
        self.assertIsInstance(predicted_proba_array, np.ndarray)
        self.assertEqual(predicted_value_array.shape, (len(TEST_SMILES),))
        self.assertEqual(predicted_proba_array.shape, (len(TEST_SMILES), 2))


if __name__ == "__main__":
    unittest.main()



"""Classes to transform given input to a RDKit molecule."""

from __future__ import annotations

from typing import Any, Optional

from molpipeline.abstract_pipeline_elements.core import (
    AnyToMolPipelineElement,
    InvalidInstance,
)
from molpipeline.any2mol.bin2mol import BinaryToMol
from molpipeline.any2mol.inchi2mol import InchiToMol
from molpipeline.any2mol.sdf2mol import SDFToMol
from molpipeline.any2mol.smiles2mol import SmilesToMol
from molpipeline.utils.molpipeline_types import OptionalMol, RDKitMol


class AutoToMol(AnyToMolPipelineElement):
    """Transforms various inputs to RDKit Mol objects.

    A cascade of if clauses is tried to transformer the given input to a molecule.
    """

    elements: tuple[AnyToMolPipelineElement, ...]

    def __init__(
        self,
        name: str = "auto2mol",
        n_jobs: int = 1,
        uuid: Optional[str] = None,
        elements: tuple[AnyToMolPipelineElement, ...] = (
            SmilesToMol(),
            InchiToMol(),
            BinaryToMol(),
            SDFToMol(),
        ),
    ) -> None:
        """Initialize AutoToMol.

        Parameters
        ----------
        name: str, optional (default="auto2mol")
            Name of PipelineElement
        n_jobs: int, optional (default=1")
            Number of parallel jobs to use.
        uuid: str, optional (default=None)
            Unique identifier of PipelineElement.
        elements: tuple[AnyToMol, ...], optional (default=(SmilesToMol(),
            InchiToMol(), BinaryToMol(), SDFToMol()))
            Elements to try to transform the input to a molecule.
        """
        super().__init__(name=name, n_jobs=n_jobs, uuid=uuid)
        # pipeline elements for transforming the input to a molecule
        self.elements = elements

    def pretransform_single(self, value: Any) -> OptionalMol:
        """Transform input value to molecule.

        Parameters
        ----------
        value: Any
            Input value.

        Returns
        -------
        OptionalMol
            Rdkit molecule if the input can be transformed, else None.
        """
        if value is None:
            return InvalidInstance(
                self.uuid,
                f"Invalid input molecule: {value}",
                self.name,
            )

        if isinstance(value, RDKitMol):
            return value

        # sequentially try to transform the input to a molecule using predefined elements
        for element in self.elements:
            mol = element.pretransform_single(value)
            if not isinstance(mol, InvalidInstance):
                return mol

        return InvalidInstance(
            self.uuid,
            f"Not readable input molecule: {value}",
            self.name,
        )



"""Abstract classes for transforming rdkit molecules to float vectors."""

# pylint: disable=too-many-arguments

from __future__ import annotations

import abc
from typing import Any, Iterable, Optional, Union

try:
    from typing import Self  # type: ignore[attr-defined]
except ImportError:
    from typing_extensions import Self

import numpy as np
import numpy.typing as npt
from sklearn.base import clone
from sklearn.preprocessing import StandardScaler

from molpipeline.abstract_pipeline_elements.core import (
    InvalidInstance,
    MolToAnyPipelineElement,
)
from molpipeline.utils.molpipeline_types import AnyTransformer, RDKitMol


class MolToDescriptorPipelineElement(MolToAnyPipelineElement):
    """PipelineElement which generates a matrix from descriptor-vectors of each molecule."""

    _standardizer: Optional[AnyTransformer]
    _output_type = "float"
    _feature_names: list[str]

    def __init__(
        self,
        standardizer: Optional[AnyTransformer] = StandardScaler(),
        name: str = "MolToDescriptorPipelineElement",
        n_jobs: int = 1,
        uuid: Optional[str] = None,
    ) -> None:
        """Initialize MolToDescriptorPipelineElement.

        Parameters
        ----------
        standardizer: Optional[AnyTransformer], default=StandardScaler()
            The output is post_processed according to the standardizer if not None.
        name: str:
            Name of the PipelineElement.
        n_jobs: int:
            Number of jobs to use for parallelization.
        uuid: Optional[str]
            UUID of the PipelineElement.

        Returns
        -------
        None
        """
        super().__init__(name=name, n_jobs=n_jobs, uuid=uuid)
        self._standardizer = standardizer
        if self._standardizer is not None:
            self._requires_fitting = True
        self._mean = None
        self._std = None

    @property
    @abc.abstractmethod
    def n_features(self) -> int:
        """Return the number of features."""

    @property
    def feature_names(self) -> list[str]:
        """Return a copy of the feature names."""
        return self._feature_names[:]

    def assemble_output(
        self,
        value_list: Iterable[npt.NDArray[np.float64]],
    ) -> npt.NDArray[np.float64]:
        """Transform output of all transform_single operations to matrix.

        Parameters
        ----------
        value_list: Iterable[npt.NDArray[np.float64]]
            List of numpy arrays with calculated descriptor values of each molecule.

        Returns
        -------
        npt.NDArray[np.float64]
            Matrix with descriptor values of each molecule.
        """
        return np.vstack(list(value_list))

    def get_params(self, deep: bool = True) -> dict[str, Any]:
        """Return all parameters defined during object initialization.

        Parameters
        ----------
        deep: bool, default=True
            If True get a deep copy of the parameters.

        Returns
        -------
        dict[str, Any]
            Dictionary containing all parameters relevant to initialize the object with same properties.
        """
        params = super().get_params(deep)
        if deep:
            if self._standardizer is not None:
                params["standardizer"] = clone(self._standardizer)
            else:
                params["standardizer"] = None
        else:
            params["standardizer"] = self._standardizer
        return params

    def set_params(self, **parameters: Any) -> Self:
        """Set parameters.

        Parameters
        ----------
        parameters: Any
            Dictionary with parameter names and corresponding values.

        Returns
        -------
        Self
            Object with updated parameters.
        """
        parameter_copy = dict(parameters)
        standardizer = parameter_copy.pop("standardizer", None)
        if standardizer is not None:
            self._standardizer = standardizer
        super().set_params(**parameter_copy)
        return self

    def fit_to_result(self, values: list[npt.NDArray[np.float64]]) -> Self:
        """Fit object to data.

        Parameters
        ----------
        values: list[RDKitMol]
            List of RDKit molecules to which the Pipeline element is fitted.

        Returns
        -------
        Self
            Fitted MolToDescriptorPipelineElement.
        """
        value_matrix = np.vstack(list(values))
        if self._standardizer is not None:
            self._standardizer.fit(value_matrix, None)
        return self

    def _normalize_matrix(
        self, value_matrix: npt.NDArray[np.float64]
    ) -> npt.NDArray[np.float64]:
        """Normalize matrix with descriptor values.

        Parameters
        ----------
        value_matrix: npt.NDArray[np.float64]
            Matrix with descriptor values of molecules.

        Returns
        -------
        npt.NDArray[np.float64]
            Normalized matrix with descriptor values of molecules.
        """
        if self._standardizer is not None:
            return self._standardizer.transform(value_matrix)
        return value_matrix

    def transform(self, values: list[RDKitMol]) -> npt.NDArray[np.float64]:
        """Transform the list of molecules to sparse matrix.

        Parameters
        ----------
        values: list[RDKitMol]
            List of RDKit molecules for which the descriptor vectors are calculated.

        Returns
        -------
        npt.NDArray[np.float64]
            Matrix with descriptor values of molecules.
        """
        descriptor_matrix: npt.NDArray[np.float64] = super().transform(values)
        return descriptor_matrix

    def finalize_single(
        self, value: npt.NDArray[np.float64]
    ) -> npt.NDArray[np.float64]:
        """Finalize single value. Here: standardize vector.

        Parameters
        ----------
        value: Any
            Single value to be finalized.

        Returns
        -------
        Any
            Finalized value.
        """
        if self._standardizer is not None:
            standadized_value = self._standardizer.transform(value.reshape(1, -1))
            return standadized_value.reshape(-1)
        return value

    @abc.abstractmethod
    def pretransform_single(
        self, value: RDKitMol
    ) -> Union[npt.NDArray[np.float64], InvalidInstance]:
        """Transform mol to dict, where items encode columns indices and values, respectively.

        Parameters
        ----------
        value: Chem.Mol
            Molecule to be transformed to descriptor vector.

        Returns
        -------
        npt.NDArray[np.float64]
            Vector with descriptor values of molecule.
        """
